/**
 * \file application.cpp
 * QApplication with settings.
 *
 * \b Project: Qoobar
 * \author Alex Novichkov
 * \date 17 May 2010
 *
 * Copyright (C) 2010  Alex Novichkov
 *
 * This file is part of Qoobar.
 *
 * Qoobar is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Qoobar is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "application.h"
#ifdef HAVE_QT5
#include <QtWidgets>
#include <QtConcurrent/QtConcurrent>
#else
#include <QtGui>
#endif
#include "id3v1stringhandler.h"

#ifdef Q_OS_LINUX
#include "linuxpaths.h"
#endif
#ifdef Q_OS_WIN
#include "windowspaths.h"
#endif
#ifdef Q_OS_MAC
#include "macpaths.h"
#endif
#ifdef Q_OS_OS2
#include "os2paths.h"
#endif

#include "iqoobarplugin.h"
#include "idownloadplugin.h"
#include "enums.h"
#include "qoobarglobals.h"

Application::Application(int &argc, char **argv, bool useGui)
    :  QApplication(argc,argv,useGui)
{
    if (useGui) setWindowIcon(QIcon(QSL(":/src/icons/qoobar.png")));
    appTranslator = new QTranslator(this);
    qtTranslator = new QTranslator(this);
    installTranslator(appTranslator);
    installTranslator(qtTranslator);

    consoleMode = false;


    paths = 0;
#ifdef Q_OS_LINUX
    paths = new LinuxPaths();
#endif
#ifdef Q_OS_WIN
    paths = new WindowsPaths();
#endif
#ifdef Q_OS_MAC
    paths = new MacPaths();
#endif
#ifdef Q_OS_OS2
    paths = new OS2Paths();
#endif

    currentScheme = 0;


    /*Testing for loadable libraries*/
#ifdef Q_OS_LINUX
    QFileInfoList libFiles = QDir(QSL("/usr/lib")).entryInfoList(QStringList()<<QSL("*discid*"), QDir::Files | QDir::NoSymLinks);
    libFiles << QDir(QSL("/usr/lib/i386-linux-gnu")).entryInfoList(QStringList()<<QSL("*discid*"), QDir::Files | QDir::NoSymLinks);
    if (!libFiles.isEmpty()) discidLibraryPath = libFiles.at(0).canonicalFilePath();
#else
    discidLibraryPath = QSL("libdiscid");
#endif
#ifndef Q_OS_MAC
    QLibrary lib(discidLibraryPath);
    if (!lib.load())
        discidLibraryPath.clear();
    lib.unload();
#endif
}

template <class T>
QVariantList makeList(const QVector<T> &list)
{
    QVariantList result;
    Q_FOREACH(const T &t,list) result.append(QVariant(t));
    return result;
}

Application::~Application()
{
    if (!consoleMode) {
        writeGlobalSettings();
        writeGuiSettings();
    }
    delete currentScheme;
    currentScheme=0;
}

void Application::loadTranslations()
{
    Q_ASSERT(paths);
    if (qtTranslator)
        qtTranslator->load(QSL("qt_") + langID, QLibraryInfo::location(QLibraryInfo::TranslationsPath));
    if (appTranslator)
        appTranslator->load(paths->translationsPath()+QSL("/qoobar_") + langID);
}

QStringList Application::readResource(const QString &s)
{
    QString fileName = QString("%1/%2.txt").arg(paths->userCompletionsPath()).arg(s);
    if (!QFileInfo(fileName).exists())
        fileName = QString("%1/%2.txt").arg(paths->completionsPath()).arg(s);
    QFile file(fileName);
    if (file.open(QFile::ReadOnly | QFile::Text)) {
        QTextStream in(&file);
        in.setCodec("UTF-8");
        return in.readAll().split('\n');
    }
    return QStringList();
}

void Application::writeResources()
{
    QString path = paths->userCompletionsPath();
    if (!QDir().exists(path))
        if (!QDir().mkpath(path)) {
            QMessageBox::critical(0,tr("Qoobar"),tr("Cannot write resource files into\n%1.\n"
                                                    "Please check the folder").arg(path));
            return;
        }


    for (int tagID=0; tagID<TAGSCOUNT; ++tagID) {
        QStringList list=completions.value(tagID);
        QFile file(QString("%1/%2.txt").arg(path).arg(untranslatedTagByID(tagID)));
        if (!list.isEmpty()) {
            if (file.open(QFile::WriteOnly | QFile::Text)) {
                QTextStream out(&file);
                out.setCodec("UTF-8");
                out << list.join(QChar('\n'));
            }
            else {
                QMessageBox::critical(0,tr("Qoobar"),tr("Cannot write %1.\n"
                                                        "Please check the file").arg(file.fileName()));
                return;
            }
        }
        else
            file.remove();
    }
}

void Application::setId3v1Encoding(const QString &s)
{
    id3v1Encoding=s;
    ID3v1StringHandler *stringHandler = new ID3v1StringHandler;
    stringHandler->setCodec(QTextCodec::codecForName(id3v1Encoding.toLatin1()));
}

void Application::updateCompletion(const int tagID, const QString &appended)
{
    if (appended.isEmpty()) return;
    useCompletion[tagID] = true;

    QStringList list=completions.value(tagID);
    if (!list.contains(appended))
        list.append(appended);
    completions[tagID]=list;
}

bool keepNotEmpty(const QString &s)
{
    return !s.isEmpty();
}

void Application::updateCompletions(const int tagID, const QStringList &collected)
{
    useCompletion[tagID] = true;
    QStringList list=completions.value(tagID);
    list.append(collected);
    list.removeDuplicates();
    QtConcurrent::blockingFilter(list,keepNotEmpty);
    completions[tagID]=list;
}

void Application::readGuiSettings()
{
    QSettings se(QSL("qoobar"),QSL("gui"));

    //Interface
    useUndo = se.value(QSL("useUndo"),false).toBool();
    autoexpand=se.value(QSL("autoexpand"),true).toBool();
    static const QChar rawChars[] = {0xdf,
        0xe0, 0xe1, 0xe2, 0xe3, 0xe4, 0xe5, 0xe6, 0xe7, 0xe8, 0xe9, 0xea, 0xeb, 0xec, 0xed, 0xee, 0xef,
        0xf0, 0xf1, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6,       0xf8, 0xf9, 0xfa, 0xfb, 0xfc, 0xfd, 0xfe, 0xff,
             0x101,      0x103,      0x105,      0x107,      0x109,      0x10b,      0x10d,      0x10f,
             0x111,      0x113,      0x115,      0x117,      0x119,      0x11b,      0x11d,      0x11f,
             0x121,      0x123,      0x125,      0x127,      0x129,      0x12b,      0x12d,      0x12f,
             0x131,      0x133,      0x135,      0x137,0x138,      0x13a,      0x13c,      0x13e,
       0x140,      0x142,      0x144,      0x146,      0x148,0x149,      0x14b,      0x14d,      0x14f,
             0x151,      0x153,      0x155,      0x157,      0x159,      0x15b,      0x15d,      0x15f,
             0x161,      0x163,      0x165,      0x167,      0x169,      0x16b,      0x16d,      0x16f,
             0x171,      0x173,      0x175,      0x177,            0x17a,      0x17c,      0x17e
    };

    chars=se.value(QSL("chars"),QString(rawChars,96)).toString();
    QVariant charsFontV=se.value(QSL("charsFont"),font());
    charsFont=charsFontV.value<QFont>();

    langID=se.value(QSL("lang"),QSL("en")).toString();
    loadTranslations();

    player=se.value(QSL("player"),QSL("")).toString();
    saveChanges=se.value(QSL("saveChanges"),false).toBool();

    patterns=se.value(QSL("patterns"),QSL("[%N. ]%t;%a/[%N. ]%t;[%N. ]%A - %t")).toString().split(';');
    QString pattern=se.value(QSL("pattern")).toString();
    if (!pattern.isEmpty()) {
        addRenamePattern(pattern);
        se.remove(QSL("pattern"));
    }

    lastDirectory=se.value(QSL("lastDirectory"),QDir::homePath()).toString();
    splitterState = se.value(QSL("splitter")).toByteArray();
    dirSplitterState = se.value(QSL("dirSplitter")).toByteArray();
    showDirView = se.value(QSL("showDirView"),true).toBool();

    freedbHello = se.value(QSL("freedbHello"),
                           QString("&hello=novichkov+qoobar.sourceforge.net+qoobar+%1&proto=6")
                           .arg(QOOBAR_VERSION)).toString();
    se.setValue(QSL("freedbHello"),freedbHello);
    useProxy = se.value(QSL("useProxy"),false).toBool();
    proxyServer = se.value(QSL("proxyServer")).toString();
    proxyPort = se.value(QSL("proxyPort")).toString();
    proxyLogin = se.value(QSL("proxyLogin")).toString();
    proxyPassword = se.value(QSL("proxyPassword")).toString();
    cdromDevice=se.value(QSL("cdromDevice")).toString();

    lastCueFile = se.value(QSL("lastCueFile")).toString();

    columns160 = se.value(QSL("columns160")).toByteArray();
    rows=se.value(QSL("rows")).toByteArray();

    tagStatus = QVector<int>(TAGSCOUNT,ALWAYSVISIBLE);
    QVariantList variantList = se.value(QSL("tagStatus")).toList();
    if (!variantList.isEmpty()) {
        for (int i=0; i<variantList.size(); ++i)
            tagStatus[i] = variantList.at(i).toInt();
    }

/** Autocompletion options*/
    //completion style:
    //    true - matches just the last word
    //    false - matches the whole string
    completionStyle = se.value(QSL("completionStyle"),true).toBool();
    //collectSilently:
    //    true - collects all text that was saved in tags
    //    false - user manually adds text to completion list
    collectSilently = se.value(QSL("collectSilently"),false).toBool();

    useCompletion = QVector<bool>(TAGSCOUNT,false);
    useCompletion[COMPOSER]=true;
    useCompletion[GENRE]=true;
    useCompletion[MOOD]=true;
    variantList = se.value(QSL("useCompletion")).toList();
    for (int i=0; i<variantList.size(); ++i)
        useCompletion[i]=variantList.at(i).toBool();

    for (int tagID=0; tagID<TAGSCOUNT; ++tagID)
        completions[tagID] = readResource(untranslatedTagByID(tagID));

    charsShown = se.value(QSL("charsShown"), true).toBool();
//    useDelayedUpdate = se.value(QSL("useDelayedUpdate"),false).toBool();
    geometry = se.value(QSL("geometry")).toByteArray();
    fillPatterns = se.value(QSL("fillPatterns"),QSL("%n. %A - %a")).toString().split(';');
}

void Application::readGlobalSettings()
{
#ifdef Q_OS_LINUX
    QSettings se(QString("%1/.config/qoobar/global.ini").arg(QDir::homePath()), QSettings::IniFormat);
#else
#ifdef Q_OS_MAC
    QSettings se(QSL("qoobar"),QSL("global"));
#else
    QSettings se(QSL("qoobar"),QSL("gui"));
#endif
#endif
    trim=se.value(QSL("simplify-whitespaces"),QSL("on")).toString()=="on";

    mp3readape=se.value(QSL("read-ape"), QSL("off")).toString()=="on";
    mp3writeape=se.value(QSL("write-ape"), QSL("off")).toString()=="on";
    mp3readid3=se.value(QSL("read-id3v2"), QSL("on")).toString()=="on";
    mp3writeid3=se.value(QSL("write-id3v2"), QSL("on")).toString()=="on";

    QString id3v1 = se.value(QSL("id3v1-behavior"), QSL("delete")).toString();
    id3v1Synchro = ID3V1_DELETE;
    if (id3v1=="update-always") id3v1Synchro = ID3V1_UPDATE_ALWAYS;
    else if (id3v1=="update-existing") id3v1Synchro = ID3V1_UPDATE_ONLY_EXISTING;

    id3v1Transliterate = se.value("id3v1-transliterate", "on").toString()=="on";
    setId3v1Encoding(se.value("id3v1-encoding","ISO-8859-1").toString());

    oggPictureNew = se.value("ogg-image-new","on").toString()=="on";

    /** rename options*/
    destinationFolder = se.value("output-dir","").toString();
    replaceWinChars   = se.value("replace-win-chars", "on").toString()=="on";
    winCharsReplacer  = se.value("win-chars-replacer","_").toString();
    replaceSpaces     = se.value("replace-spaces", "off").toString()=="on";
    spacesReplacer    = se.value("spaces-replacer","_").toString();
    removeDiacritics  = se.value("remove-diacritics", "off").toString()=="on";

    QString s = se.value("change-case", "dont-change").toString();
    changeCase = false;
    ccase = Case::FirstCaps;
    if (s=="lower") {
        changeCase = true;
        ccase = Case::AllSmall;
    }
    else if (s=="upper") {
        changeCase = true;
        ccase = Case::AllCaps;
    }
    else if (s=="first-up") {
        changeCase = true;
        ccase = Case::FirstCaps;
    }
    else if (s=="every-first-up") {
        changeCase = true;
        ccase = Case::EveryFirstCaps;
    }

    int length = se.value("maximum-length", 0).toInt();
    if (length<=0) {
        maximumLength = 127;
        trimFileLength = false;
    }
    else {
        maximumLength = length;
        trimFileLength = true;
    }

    renamingOperation = se.value("rename-operation", "rename").toString()=="rename"?0:1;
    removeFolder      = se.value("remove-folder", "on").toString()=="on";
    applyToFolders    = se.value("apply-to-folders", "off").toString()=="on";

    currentSchemeName = se.value("scheme", "default").toString();
    if (!currentScheme)
        currentScheme = new TaggingScheme(currentSchemeName);
    else
        currentScheme->setFilePath(currentSchemeName);
    currentScheme->read();

    encaGuessLanguage  = se.value("enca-guess-language","russian").toString();
    writeFieldsSeparately  = se.value("write-fields-separately","off").toString()=="on";
    copyFiles = se.value("rg-copy-files","off").toString()=="on";
    id3v2version = se.value("force-id3v23","off").toString()=="on"?3:4;

    recursive = se.value("recursive",true).toBool();
    verbose = se.value("verbose",false).toBool();

    //"header" - write rg into file header
    //"tags" - write rg into ape tags
    mpcWriteRg = se.value("mpc-replaygain-format", "header").toString()=="header";
}

void Application::writeGuiSettings()
{
    QSettings se("qoobar","gui");
    if (!se.isWritable()) {
        QMessageBox::critical(0,tr("Qoobar"),tr("Cannot write settings. The settings file is read-only"));
        return;
    }
    se.setValue("lang",langID);
    se.setValue("chars",chars);
    se.setValue("lastDirectory",lastDirectory);
    se.setValue("player",player);
    se.setValue("saveChanges",saveChanges);
    se.setValue("patterns",patterns.join(";"));
    se.setValue("autoexpand",autoexpand);
    se.setValue("splitter",splitterState);
    se.setValue("dirSplitter",dirSplitterState);
    se.setValue("showDirView",showDirView);
    se.setValue("useProxy",useProxy);
    se.setValue("proxyServer",proxyServer);
    se.setValue("proxyPort",proxyPort);
    se.setValue("proxyLogin",proxyLogin);
    se.setValue("proxyPassword",proxyPassword);
    se.setValue("cdromDevice",cdromDevice);
    se.setValue("charsFont",charsFont);
    se.setValue("lastCueFile",lastCueFile);
    se.setValue("columns160",columns160);
    se.setValue("rows",rows);
    se.setValue("tagStatus",QVariant(makeList(tagStatus)));
    se.setValue("completionStyle",completionStyle);
    se.setValue("collectSilently",collectSilently);
    se.setValue("useCompletion",QVariant(makeList(useCompletion)));
    se.setValue("charsShown", charsShown);
    se.setValue("useUndo",useUndo);
//    se.setValue("useDelayedUpdate",useDelayedUpdate);
    se.setValue("geometry",geometry);
    se.setValue("fillPatterns",fillPatterns.join(";"));

    writeResources();
}

void Application::writeGlobalSettings()
{
#ifdef Q_OS_LINUX
    QSettings se(QString("%1/.config/qoobar/global.ini").arg(QDir::homePath()), QSettings::IniFormat);
#else
#ifdef Q_OS_MAC
    QSettings se(QSL("qoobar"),QSL("global"));
#else
    QSettings se(QSL("qoobar"),QSL("gui"));
#endif
#endif
    if (!se.isWritable()) {
        qCritical("Cannot write settings. The settings file is read-only");
        return;
    }


    se.setValue("simplify-whitespaces", trim?"on":"off");
    se.setValue("read-ape", mp3readape?"on":"off");
    se.setValue("write-ape", mp3writeape?"on":"off");
    se.setValue("read-id3v2",mp3readid3?"on":"off");
    se.setValue("write-id3v2", mp3writeid3?"on":"off");

    if (id3v1Synchro == ID3V1_DELETE) se.setValue("id3v1-behavior", "delete");
    else if (id3v1Synchro == ID3V1_UPDATE_ALWAYS) se.setValue("id3v1-behavior", "update-always");
    else if (id3v1Synchro == ID3V1_UPDATE_ONLY_EXISTING) se.setValue("id3v1-behavior", "update-existing");

    se.setValue("id3v1-transliterate", id3v1Transliterate?"on":"off");
    se.setValue("id3v1-encoding", id3v1Encoding);
    se.setValue("ogg-image-new", oggPictureNew?"on":"off");

    /** rename options*/
    se.setValue("output-dir", destinationFolder);
    se.setValue("replace-win-chars",replaceWinChars? "on":"off");
    se.setValue("win-chars-replacer", winCharsReplacer);
    se.setValue("replace-spaces", replaceSpaces? "on":"off");
    se.setValue("spaces-replacer", spacesReplacer);
    se.setValue("remove-diacritics", removeDiacritics? "on":"off");

    if (!changeCase) se.setValue("change-case", "dont-change");
    else if (ccase == Case::AllSmall) se.setValue("change-case", "lower");
    else if (ccase == Case::AllCaps) se.setValue("change-case", "upper");
    else if (ccase == Case::FirstCaps) se.setValue("change-case", "first-up");
    else if (ccase == Case::EveryFirstCaps) se.setValue("change-case", "every-first-up");

    if (!trimFileLength) se.setValue("maximum-length", 0);
    else se.setValue("maximum-length", maximumLength);

    se.setValue("rename-operation", renamingOperation==0?"rename":"copy");
    se.setValue("remove-folder", removeFolder?"on":"off");
    se.setValue("apply-to-folders", applyToFolders?"on":"off");
    se.setValue("scheme", currentSchemeName);
    se.setValue("enca-guess-language", encaGuessLanguage);
    se.setValue("write-fields-separately", writeFieldsSeparately?"on":"off");
    se.setValue("rg-copy-files", copyFiles?"on":"off");
    se.setValue("force-id3v23", id3v2version==3?"on":"off");
    se.setValue("recursive",recursive);
    se.setValue("verbose",verbose);
    se.setValue("mpc-replaygain-format", mpcWriteRg ? "header" : "tags");
}

void Application::clearSettings()
{
#ifdef Q_OS_LINUX
    QSettings se(QString("%1/.config/qoobar/global.ini").arg(QDir::homePath()), QSettings::IniFormat);
    se.clear();
#endif
#ifdef Q_OS_MAC
    QSettings se(QSL("qoobar"),QSL("global"));
    se.clear();
#endif
    QSettings se1("qoobar","gui");
    se1.clear();
}

void Application::resetSettings()
{
    clearSettings();
    readGlobalSettings();
    readGuiSettings();
}

void Application::addRenamePattern(const QString &s)
{
    patterns.prepend(s);
    patterns.removeDuplicates();
}

void Application::addFillPattern(const QString &s)
{
    fillPatterns.prepend(s);
    fillPatterns.removeDuplicates();
}

void Application::loadPlugins()
{
    QDir pluginsDir = QDir(paths->pluginsPath());

#ifdef HAVE_QT5
    Q_FOREACH (const QString &fileName, pluginsDir.entryList(QDir::Files)) {
        QString path=pluginsDir.absoluteFilePath(fileName);
        QPluginLoader loader(path);
        QJsonObject metaData = loader.metaData().value("MetaData").toObject();

        if (metaData.isEmpty()) continue;

        metaData.insert("path",path);
        QString pluginInterface=metaData.value(QSL("interface")).toString();
        if (pluginInterface=="IDownloadPlugin")
            downloadPlugins << metaData;
        if (pluginInterface=="IQoobarPlugin")
            plugins << metaData;
    }
#else
    Q_FOREACH (const QString &fileName, pluginsDir.entryList(QDir::Files)) {
        QString path=pluginsDir.absoluteFilePath(fileName);
        QPluginLoader loader(path);
        QObject *plugin = loader.instance();
        if (plugin) {
            IQoobarPlugin *iplugin = qobject_cast<IQoobarPlugin *>(plugin);
            if (iplugin) plugins.insert(iplugin->key(),iplugin);
            IDownloadPlugin *iplugin1 = qobject_cast<IDownloadPlugin *>(plugin);
            if (iplugin1) downloadPlugins.insert(iplugin1->key(),iplugin1);
        }
    }
#endif

}
