/**
 * \file argsparser.cpp
 * Arguments parser and command line processor
 *
 * \b Project: Qoobar
 * \author Alex Novichkov
 * \date 10 Apr 2013
 *
 * Copyright (C) 2013  Alex Novichkov
 *
 * This file is part of Qoobar.
 *
 * Qoobar is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Qoobar is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "argsparser.h"


#include "CmdLine.h"
#include "Visitor.h"
#include "qoobarglobals.h"
#include "model.h"
#include "enums.h"
#include "placeholders.h"
#include "application.h"
#include "replaygainer.h"
#include "tagparser.h"
#include "stringroutines.h"
#include "tagrender.h"
#include "mainwindow.h"
#include "cuesplitter.h"

//#include <QtDebug>
#include <QFile>
#include <QImage>

const char *keys[TAGSCOUNT] = {
    "COMPOSER",
    "ALBUM",
    "TITLE",
    "PERFORMER",
    "ARTIST",
    "CONDUCTOR",
    "ORCHESTRA",
    "SUBTITLE",
    "KEY",
    "COMMENT",
    "GENRE",
    "YEAR",
    "TRACKNUMBER",
    "TOTALTRACKS",
    "ALBUMARTIST",
    "CATEGORY",
    "PUBLISHER",
    "COPYRIGHT",
    "MOOD",
    "TEMPO",
    "LYRICIST",
    "LYRICS",
    "DISCNUMBER",
    "TOTALDISCS",
    "ENCODEDBY",
    "REMIXEDBY",
    "RATING",
    "ORIGINALALBUM",
    "ORIGINALARTIST",
    "ORIGINALLYRICIST",
    "URL",
    "ISRC"
};

int getRow(const QString &key)
{
    QString s=key.toUpper();
    for (int i=0; i < TAGSCOUNT; ++i) {
        if (s==keys[i]) return i;
    }
    return -1;
}

const char *imageTypes[21]= {
    "OTHER",
    "ICON",
    "OTHER_ICON",
    "FRONT_COVER",
    "BACK_COVER",
    "LEAFLET",
    "MEDIA",
    "LEAD_ARTIST",
    "ARTIST",
    "CONDUCTOR",
    "BAND",
    "COMPOSER",
    "LYRICIST",
    "RECORDING_LOCATION",
    "DURING_RECORDING",
    "DURING_PERFORMANCE",
    "VIDEO",
    "BRIGHT_COLORED_FISH",
    "ILLUSTRATION",
    "BAND_LOGO",
    "PUBLISHER_LOGO"
};

const char* fileTypes[] = {
    "Mpeg Layer III",
    "OGG file",
    "Windows Media Audio",
    "FLAC",
    "M4A",
    "Musepack",
    "WavPack",
    "WAV",
    "Speex",
    "TrueType Audio",
    "OGA",
    "Monkey's Audio",
    "AIFF",
    "OPUS"
};

class PlaceholdersVisitor : public TCLAP::Visitor
{
public:
    PlaceholdersVisitor() : TCLAP::Visitor() {}
    void visit()
    {
        std::wcout << QObject::tr("Available placeholders for --fill and --rename:").toStdWString()
                   << std::endl;
        for (int i = 0; i < Placeholders::placeholdersCount; ++i) {
            std::wcout <<"    "<<Placeholders::placeholders[i].text
                       << "    " << Placeholders::placeholders[i].description
                       << std::endl;
        }
        std::wcout  << std::endl;
        qApp->quit();
    }
};

class KeysVisitor : public TCLAP::Visitor
{
public:
    KeysVisitor() : TCLAP::Visitor() {}
    void visit()
    {
        std::wcout << QObject::tr("Available keys for --set and --remove:").toStdWString()
                   << std::endl;
        for (int i=0; i< TAGSCOUNT; ++i)
            std::wcout << "    " << keys[i] << std::endl;
        std::wcout << std::endl;
        qApp->quit();
    }
};

class ImageTypesVisitor : public TCLAP::Visitor
{
public:
    ImageTypesVisitor() : TCLAP::Visitor() {}
    void visit()
    {
        std::wcout << QObject::tr("Available image types for --set-image-type:").toStdWString()
                   << std::endl;
        for (int i=0; i<21; ++i)
            std::wcout << "    " << imageTypes[i] << std::endl;
        std::wcout << std::endl;
        qApp->quit();
    }
};

QStringList createFilesToAdd(const std::vector<QString> &filesNames)
{
    QStringList result;

    for (uint i=0; i<filesNames.size(); ++i)
        result << filesNames.at(i);
    return result;
}

QString createRows(const std::vector<QString> &keys) {
    QString s;
    for (uint i=0; i<keys.size(); ++i) {
        QString k = keys.at(i);
        for (int j=0; j<k.size(); ++j) {
            if (!s.contains(QString(k[j]))) s.append(QString(k[j]));
        }
    }
    return s;
}

ArgsParser::ArgsParser(QObject *parent) : QObject(parent)
{
}

typedef TCLAP::ValueArg<QString> StringArg;
typedef TCLAP::MultiArg<QString> StringMultiArg;
typedef TCLAP::ValueArg<uint> IntArg;
typedef TCLAP::SwitchArg SwitchArg;

bool commandLineMode(const QList<TCLAP::Arg *> &commandArgs)
{
    Q_FOREACH(TCLAP::Arg *arg, commandArgs) {
        if (arg->isSet()) {
            return true;
        }
    }
    return false;
}

void ArgsParser::parse()
{
    QStringList args = qApp->arguments();
    try {
        TCLAP::CmdLine cmdLine("Qoobar, a simple tagger for classical music", ' ', QString(QOOBAR_VERSION), true);

        //standard tags
        StringArg composerArg("c","composer","Sets composer","Changing tags", false,"","STRING");
        StringArg albumArg("a","album","Sets album", "Changing tags", false, "", "STRING");
        StringArg titleArg("t","title","Sets title", "Changing tags", false, "", "STRING");
        StringArg artistArg("A","artist","Sets artist", "Changing tags", false, "", "STRING");
        StringArg commentArg("C","comment","Sets comment", "Changing tags", false, "", "STRING");
        StringArg genreArg("g","genre","Sets genre", "Changing tags", false, "", "STRING");
        IntArg yearArg("y","year","Sets year", "Changing tags", false, 0, "NUM");
        IntArg trackNumberArg("n","track-number","Sets track number", "Changing tags", false, 0, "NUM");
        IntArg totalTracksArg("T","total-tracks","Sets total tracks", "Changing tags", false, 0, "NUM");
        StringMultiArg setArg("s","set","Sets tag with KEY to STRING. "
                              "See --list-keys for complete list of standard keys. "
                              "If KEY is not standard, the user tag with key KEY will be set",
                              "Changing tags", false, "KEY:STRING");

        //removing tags
        StringMultiArg removeArg("","remove","Removes tags by their KEY. "
                                 "See --list-keys for complete list of standard keys. "
                                 "If KEY is not standard, the user tag with key KEY will be removed",
                                 "Changing tags", false, "KEY");
        SwitchArg removeUserTagsArg("","remove-user-tags","Removes all non-standard tags", "Changing tags");
        SwitchArg removeAllArg("","remove-all","Removes all tags, including standard tags, user tags, image, and replaygain info", "Changing tags");

        //images handling
        StringArg setImageArg("i","set-image", "Sets an image to the file. If PATH is empty, image will be removed",
                              "Changing image", false, "", "PATH");
        StringArg setImageDescriptionArg("","set-image-description","Sets an image description", "Changing image", false, "", "STRING");
        std::vector<QString> allowedImageTypes;
        for (int i=0; i<21; ++i) allowedImageTypes.push_back(imageTypes[i]);
        TCLAP::ValuesConstraint<QString> allowedImageTypesVals(allowedImageTypes);
        StringArg setImageTypeArg("","set-image-type","Sets an image type. "
                                  "For list of possible types see --list-image-types",
                                  "Changing image", false, "FRONT_COVER", &allowedImageTypesVals);

        SwitchArg removeImageArg("","remove-image","Removes the image from the file", "Changing image");
        StringArg writeImageArg("","write-image","Causes the attached image to be written into DIR directory. "
                                "If DIR is empty, the image will be written where the audio file is. ",
                                "Changing image", false, "", "DIR");

        //replay gaining
        std::vector<QString> allowedRG;
        allowedRG.push_back("file");
        allowedRG.push_back("album");
        allowedRG.push_back("dir");
        allowedRG.push_back("tag");
        TCLAP::ValuesConstraint<QString> allowedRGVals(allowedRG);
        StringArg scanRgArg("","scan-rg","Scans the file without writing ReplayGain info into files.\n"
                            "        <file>: Scans per file track gain\n"
                            "        <album>: Scans all files as a single album\n"
                            "        <dir>: Scans each directory as a single album\n"
                            "        <tag>: Scans each ALBUM as a single album",
                            "ReplayGain info", false, "file", &allowedRGVals);
        StringArg writeRgArg("","write-rg","Scans the file and writes ReplayGain info into files.\n"
                            "        <file>: Scans per file track gain\n"
                            "        <album>: Scans all files as a single album\n"
                            "        <dir>: Scans each directory as a single album\n"
                            "        <tag>: Scans each ALBUM as a single album",
                            "ReplayGain info", false, "file", &allowedRGVals);
        SwitchArg removeRGArg("","remove-rg","Removes ReplayGain info from the file","ReplayGain info");
        SwitchArg skipScannedArg("","skip-scanned-rg","Switches on skipping files that already have ReplayGain info","ReplayGain info");
        std::vector<QString> allowedmpcOptions;
        allowedmpcOptions.push_back("header");
        allowedmpcOptions.push_back("tags");
        TCLAP::ValuesConstraint<QString> mpcoptions(allowedmpcOptions);
        StringArg mpcReplaygainFormatArg("","mpc-replaygain-format","Sets the format of ReplayGain information "
                                         "to be written into Musepack files.\n"
                                         "        <header> - ReplayGain information will be written into the file header "
                                         "and cannot be changed afterwards.\n"
                                         "        <tags> - ReplayGain information will be written into APEv2 tags and can be changed or deleted.\n"
                                 "The default value is <header>.", "Global options",false, "header", &mpcoptions);

        //filling tags
        StringArg fillArg("f","fill","Fills tags from file name by a pattern using placeholders: %A (artist), "
                          "%a (album), %t (title), etc. See --list-placeholders for complete list"
                          "of placeholders", "Filling tags",false, "", "PATTERN");

        StringArg splitArg("","split","Splits an audio file by its CUE sheet","Splitting files",false, "", "PATH");

        //renaming files
        std::vector<QString> allowedReplaceWinChars;
        allowedReplaceWinChars.push_back("on");
        allowedReplaceWinChars.push_back("off");
        TCLAP::ValuesConstraint<QString> options(allowedReplaceWinChars);
        StringArg renameArg("","rename","Rename file (the extension is not affected) based on "
                            "data in the tag using placeholders: %A (artist), %a (album), %t (title), etc.","Renaming files", false, "", "PATTERN");
        IntArg maximumLengthArg("","maximum-length","Sets maximum file name length for --rename operation "
                                "to a specified value. If value is 0, file name will not be truncated. The default value is 0",
                                "Renaming files",false, 0, "NUM");
        StringArg replaceWinCharsArg("","replace-win-chars","Switches on/off Windows chars replacement for --rename operation. "
                                     "The default value is \"on\"","Renaming files",false, "on", &options);
        StringArg outputDirArg("","output-dir","Specifies the destination directory for --rename operation. "
                               "If <DIR> is empty, the source directory will be used.","Renaming files",false, "", "DIR");
        StringArg replaceSpacesArg("","replace-spaces","Switches on/off spaces replacement for --rename operation. "
                                   "The default value is \"off\"","Renaming files",false, "off", &options);
        StringArg removeDiacriticsArg("","remove-diacritics","Switches on/off diacritics removing for --rename operation. "
                                      "The default value is \"off\"","Renaming files",false, "off", &options);
        StringArg applyToFoldersArg("","apply-to-folders", "If this option is set, all other options for --rename operation"
                                    " will take place for created folders names as well. The default value is \"off\""
                                    ,"Renaming files",false, "off", &options);
        std::vector<QString> allowedCases;
        allowedCases.push_back("lower");
        allowedCases.push_back("upper");
        allowedCases.push_back("first-up");
        allowedCases.push_back("every-first-up");
        allowedCases.push_back("dont-change");
        TCLAP::ValuesConstraint<QString> caseOptions(allowedCases);
        StringArg changeCaseArg("","change-case","If this option is set, created file and "
                                "folders names will be in case specified by this option. "
                                "The default value is \"dont-change\"","Renaming files",false, "dont-change", &caseOptions);


        TCLAP::UnlabeledMultiArg<QString> filesArg("files", "Files and directories to process", "", false ,"PATH");

        //options
        SwitchArg recursiveArg("r","recursive","If set, directories to process are added recursively, "
                               "that is with all their subdirectories. The default behavior is recursive","Global options");
        StringArg simplifyWhitespacesArg("","simplify-whitespaces","Switches on/off tags simplifying when saving files. "
                                         "The default value is \"on\"","Global options",false, "on", &options);
        StringArg readId3v2Arg("","read-id3v2","Switches on/off reading of id3v2 tags in mp3 files. "
                               "The default value is \"on\"","Global options",false, "on", &options);
        StringArg readApeArg("","read-ape","Switches on/off reading of ape tags in mp3 files. "
                             "The default value is \"off\"","Global options",false, "off", &options);
        StringArg writeId3v2Arg("","write-id3v2","Switches on/off writing of id3v2 tags in mp3 files. "
                                "The default value is \"on\"","Global options",false, "on", &options);
        StringArg writeApeArg("","write-ape","Switches on/off writing of ape tags in mp3 files. "
                              "The default value is \"off\"","Global options",false, "off", &options);
        StringArg forceId3v23Arg("","force-id3v23","Switches on/off writing of id3v2 tags to id3v2.3 format. "
                                 "The default value is \"off\"","Global options",false, "off", &options);
        std::vector<QString> allowedid3v1Options;
        allowedid3v1Options.push_back("update-always");
        allowedid3v1Options.push_back("update-existing");
        allowedid3v1Options.push_back("delete");
        TCLAP::ValuesConstraint<QString> id3v1Options(allowedid3v1Options);
        StringArg deleteId3v1Arg("","id3v1-behavior","Switches on/off deleting of id3v1 tags when saving mp3 files. "
                                 "The default value is \"delete\"","Global options",false, "delete", &id3v1Options);
        StringArg id3v1TransliterateArg("","id3v1-transliterate","Switches on/off transliteration of id3v1 tags when saving mp3 files. "
                                        "The default value is \"on\"","Global options",false, "on", &options);
        StringArg oggImageNewArg("","ogg-image-new","Switches on/off new ogg image format. "
                                 "The default value is \"on\"","Global options",false, "on", &options);
        SwitchArg verboseArg("","verbose", "Switches on verbose mode, in which all messages are shown. "
                             "By default only errors and warnings are shown", "");
//        SwitchArg dryRunArg("d","dry-run", "Switches on dry-run mode, in which no actual changing of tags"
//                            "will take place", "");

        //LISTS AND INFO
        SwitchArg listImageTypesArg("","list-image-types","Lists all possible image types and exits", "", false, new ImageTypesVisitor());
        SwitchArg listKeysArg("","list-keys","Lists all possible tags keys and exits", "", false, new KeysVisitor());
        SwitchArg listPlaceholdersArg("","list-placeholders","Lists all possible placeholders and exits", "", false, new PlaceholdersVisitor());
        SwitchArg showTagsArg("","show-tags","Shows all tags of files", "");

        //quiet
        //apply-to
        //to-v2.3
        //to-v2.4
        //apply-function case simplify remove-diacritics transliterate fix-encoding


        QList<TCLAP::Arg *> commandArgs;


        cmdLine.add(composerArg);          cmdLine.add(albumArg);
        cmdLine.add(titleArg);             cmdLine.add(artistArg);
        cmdLine.add(commentArg);           cmdLine.add(genreArg);
        cmdLine.add(yearArg);              cmdLine.add(trackNumberArg);
        cmdLine.add(totalTracksArg);       cmdLine.add(setArg);
        cmdLine.add(removeArg);            cmdLine.add(removeUserTagsArg);
        cmdLine.add(removeAllArg);         cmdLine.add(setImageArg);
        cmdLine.add(setImageDescriptionArg);  cmdLine.add(setImageTypeArg);
        cmdLine.add(removeImageArg);       cmdLine.add(writeImageArg);
        cmdLine.add(scanRgArg);            cmdLine.add(writeRgArg);
        cmdLine.add(removeRGArg);          cmdLine.add(skipScannedArg);
        cmdLine.add(mpcReplaygainFormatArg);
        cmdLine.add(listImageTypesArg);    cmdLine.add(listKeysArg);
        cmdLine.add(listPlaceholdersArg);  cmdLine.add(showTagsArg);
        cmdLine.add(fillArg);              cmdLine.add(renameArg);
        cmdLine.add(splitArg);
        //replace options
        cmdLine.add(maximumLengthArg);
        cmdLine.add(replaceWinCharsArg);  cmdLine.add(outputDirArg);
        cmdLine.add(replaceSpacesArg);    cmdLine.add(removeDiacriticsArg);
        cmdLine.add(applyToFoldersArg);   cmdLine.add(changeCaseArg);
        //global options
        cmdLine.add(recursiveArg);
        cmdLine.add(simplifyWhitespacesArg);
        cmdLine.add(readId3v2Arg);
        cmdLine.add(readApeArg);
        cmdLine.add(writeId3v2Arg);
        cmdLine.add(writeApeArg);
        cmdLine.add(forceId3v23Arg);
        cmdLine.add(deleteId3v1Arg);
        cmdLine.add(id3v1TransliterateArg);
        cmdLine.add(oggImageNewArg);
        cmdLine.add(verboseArg);
//        cmdLine.add(dryRunArg);

        cmdLine.add(filesArg);

        commandArgs.append(&composerArg);          commandArgs.append(&albumArg);
        commandArgs.append(&titleArg);             commandArgs.append(&artistArg);
        commandArgs.append(&commentArg);           commandArgs.append(&genreArg);
        commandArgs.append(&yearArg);              commandArgs.append(&trackNumberArg);
        commandArgs.append(&totalTracksArg);       commandArgs.append(&setArg);
        commandArgs.append(&removeArg);            commandArgs.append(&removeUserTagsArg);
        commandArgs.append(&removeAllArg);         commandArgs.append(&setImageArg);
        commandArgs.append(&setImageDescriptionArg);  commandArgs.append(&setImageTypeArg);
        commandArgs.append(&removeImageArg);       commandArgs.append(&writeImageArg);
        commandArgs.append(&scanRgArg);            commandArgs.append(&removeRGArg);
        commandArgs.append(&writeRgArg);
        commandArgs.append(&skipScannedArg);       commandArgs.append(&listImageTypesArg);
        commandArgs.append(&listKeysArg);          commandArgs.append(&showTagsArg);
        commandArgs.append(&listPlaceholdersArg);  commandArgs.append(&fillArg);
        commandArgs.append(&renameArg);            commandArgs.append(&splitArg);

        std::vector<QString> vargs = args.toVector().toStdVector();
        cmdLine.parse(vargs);

        //extract file names to process in GUI mode
        QStringList filesNames;
        std::vector<int> filesIndexes = filesArg.indexes;
        Q_FOREACH (int i, filesIndexes) filesNames << args.at(i);
        files = expandFilesNames(filesNames, recursiveArg.isSet());
        App->setFilesNames(files);
        bool verbose = verboseArg.isSet();
        App->recursive = recursiveArg.isSet();
        App->verbose = verbose;
//        bool dryRun = dryRunArg.isSet();

        //First set all options for this session
        //rename options
        if (verbose) {
            appendMessage("<font color=blue>Applying options...</font>");
        }
        if (maximumLengthArg.isSet()) {
            uint option = maximumLengthArg.getValue();
            if (option==0) App->trimFileLength = false;
            else {
                App->trimFileLength = true;
                App->maximumLength = option;
            }
            if (verbose) {
                showMessage(MT_SUCCESS, "Maximum file name length is set to "+QString::number(option));
            }
        }
        if (changeCaseArg.isSet()) {
            QString option = changeCaseArg.getValue();
            if (option=="dont-change") App->changeCase=false;
            else {
                App->changeCase=true;
                if (option=="lower") App->ccase = Case::AllSmall;
                else if (option=="upper") App->ccase = Case::AllCaps;
                else if (option=="first-up") App->ccase = Case::FirstCaps;
                else App->ccase = Case::EveryFirstCaps;
            }
            if (verbose) {
                showMessage(MT_SUCCESS, "Change case option is set to \""+option+"\"");
            }
        }
        if (replaceWinCharsArg.isSet()) {
            App->replaceWinChars = replaceWinCharsArg.getValue()=="on";
            if (verbose) showMessage(MT_SUCCESS, "Windows chars replacement for --rename operation is switched "+replaceWinCharsArg.getValue());
        }
        if (outputDirArg.isSet()) {
            App->destinationFolder = outputDirArg.getValue();
            if (verbose) showMessage(MT_SUCCESS, "The destination directory for --rename operation is set to \""+outputDirArg.getValue()+"\"");
        }
        if (replaceSpacesArg.isSet()) {
            App->replaceSpaces = replaceSpacesArg.getValue()=="on";
            if (verbose) showMessage(MT_SUCCESS, "Spaces replacement for --rename operation is switched "+replaceSpacesArg.getValue());
        }
        if (removeDiacriticsArg.isSet()) {
            App->removeDiacritics = removeDiacriticsArg.getValue()=="on";
            if (verbose) showMessage(MT_SUCCESS, "Diacritics removing for --rename operation is switched "+removeDiacriticsArg.getValue());
        }
        if (applyToFoldersArg.isSet()) {
            App->applyToFolders = applyToFoldersArg.getValue()=="on";
            if (verbose)
                showMessage(MT_SUCCESS, "apply-to-folders option for --rename operation is switched "+applyToFoldersArg.getValue());
        }

        //global options
        if (deleteId3v1Arg.isSet()) {
            QString option = deleteId3v1Arg.getValue();
            int index = std::find(allowedid3v1Options.begin(),allowedid3v1Options.end(),option)-allowedid3v1Options.begin();
            if (index>=0) App->id3v1Synchro = index;
            if (verbose) showMessage(MT_SUCCESS, "Deleting of id3v1 tags is set to \""+deleteId3v1Arg.getValue()+"\"");
        }
        if (simplifyWhitespacesArg.isSet()) {
            App->trim = simplifyWhitespacesArg.getValue()=="on";
            if (verbose) showMessage(MT_SUCCESS, "Tags simplifying is switched "+simplifyWhitespacesArg.getValue());
        }
        if (readId3v2Arg.isSet()) {
            App->mp3readid3 = readId3v2Arg.getValue()=="on";
            if (verbose) showMessage(MT_SUCCESS, "Reading of id3v2 tags in mp3 files is switched "+readId3v2Arg.getValue());
        }
        if (readApeArg.isSet()) {
            App->mp3readape = readApeArg.getValue()=="on";
            if (verbose) showMessage(MT_SUCCESS, "Reading of APE tags in mp3 files is switched "+readApeArg.getValue());
        }
        if (writeId3v2Arg.isSet()) {
            App->mp3writeid3 = writeId3v2Arg.getValue()=="on";
            if (verbose) showMessage(MT_SUCCESS, "Writing of id3v2 tags in mp3 files is switched "+writeId3v2Arg.getValue());
        }
        if (writeApeArg.isSet()) {
            App->mp3writeape = writeApeArg.getValue()=="on";
            if (verbose) showMessage(MT_SUCCESS, "Writing of APE tags in mp3 files is switched "+writeApeArg.getValue());
        }
        if (forceId3v23Arg.isSet()) {
            App->id3v2version = forceId3v23Arg.getValue()=="on"?3:4;
            if (verbose) showMessage(MT_SUCCESS, "Writing of id3v2 tags to id3v2.3 format is switched "+forceId3v23Arg.getValue());
        }
        if (id3v1TransliterateArg.isSet()) {
            App->id3v1Transliterate = id3v1TransliterateArg.getValue()=="on";
            if (verbose) showMessage(MT_SUCCESS, "Transliteration of id3v1 tags is switched "+id3v1TransliterateArg.getValue());
        }
        if (oggImageNewArg.isSet()) {
            App->oggPictureNew = oggImageNewArg.getValue()=="on";
            if (verbose) showMessage(MT_SUCCESS, "Using of new ogg image format is switched "+oggImageNewArg.getValue());
        }
        if (mpcReplaygainFormatArg.isSet()) {
            App->mpcWriteRg = mpcReplaygainFormatArg.getValue()=="header";
            if (verbose) showMessage(MT_SUCCESS, "Format of Musepack Replaygain info is set to \""
                                     +mpcReplaygainFormatArg.getValue()+"\"");
        }
        if (verbose) appendMessage("");


        if (commandLineMode(commandArgs)) {
            //files can be empty, but if --split option is set,
            //we trying to add splitted files
            if (splitArg.isSet()) {
                /*if (dryRun) {
                    showMessage(MT_WARNING, "Sorry, in dry-run node file splitting is not supported.");
                }
                else*/ {
                    QString cueFile = splitArg.getValue();
                    if (verbose) {
                        appendMessage("<font color=blue>Splitting a file by \""+cueFile+"\" CUE sheet </font>");
                    }
                    if (!cueFile.isEmpty()) {
                        CueSplitter splitter;
                        connect(&splitter, SIGNAL(message(int,QString)),SLOT(showMessage(int,QString)));
                        if (verbose) connect(&splitter, SIGNAL(textReady(QString)),SLOT(addText(QString)));
                        splitter.setCueFile(cueFile);
                        splitter.split();
                        QStringList splittedFiles = splitter.splittedFiles();

                        if (!splittedFiles.isEmpty()) {
                            if (verbose)
                                showMessage(MT_SUCCESS, QString("Files splitting was successful"));
                            files.append(splittedFiles);
                        }
                    }
                    else {
                        showMessage(MT_ERROR, QString("CUE sheet file is empty!"));
                    }
                }
            }

            if (!files.isEmpty()) {
                App->setFilesNames(files);
                App->consoleMode = true;

                Model model;
                QObject::connect(&model,SIGNAL(message(int,QString)),SLOT(showMessage(int,QString)),Qt::QueuedConnection);
                model.addFiles(files);
                model.selectAll();
                if (verbose) {
                    printFiles(model.fileNames());
                }


                //next show tags and exit
                if (showTagsArg.isSet()) {
                    if (verbose) {
                        appendMessage("<font color=blue>Print all tags:</font>");
                    }
                    QList<int> indexes = model.selection();
                    Q_FOREACH (int index, indexes) {
                        Tag tag = model.fileAt(index);
                        appendMessage("<font color=blue>"+tag.fullFileName()+"</font>",75,0,5);
                        QString properties = QString("Time: %1    %2    [ %3 kbps @ %4 Hz - %5]")
                                .arg(formatLength(tag.length()))
                                .arg(fileTypes[tag.fileType()])
                                .arg(tag.bitrate())
                                .arg(tag.sampleRate())
                                .arg(tag.channels()==1?"Mono":(tag.channels()==2?"Stereo":QString("%1 channels").arg(tag.channels())));
                        appendMessage(properties,75,5,5);
                        //standard tags
                        for (int i=0; i<TAGSCOUNT; ++i) {
                            QString s = tag.tag(i);
                            if (!s.isEmpty()) {
                                appendMessage(untranslatedTagByID(i)+": "+s,75,5,5);
                            }
                        }
                        //user tags
                        QStringList ids=model.userTagsIds();
                        Q_FOREACH (const QString &id, ids) {
                            QString s = tag.userTag(id);
                            if (!s.isEmpty()) {
                                appendMessage(id+": "+s,75,5,5);
                            }
                        }
                        //image
                        if (!tag.imageIsEmpty()) {
                            QImage image;;
                            if (image.loadFromData(tag.image().pixmap())) {
                                QString format = QString("Image: %1 [ %2 x %3 ], %4")
                                        .arg(tag.image().mimetype().section("/",1))
                                        .arg(image.width())
                                        .arg(image.height())
                                        .arg(imageTypes[tag.image().type()]);
                                appendMessage(format,75,5,5);
                                if (!tag.image().description().isEmpty())
                                    appendMessage("Image description: "+tag.image().description(),75,5,5);
                            }
                        }
                        //replaygain
                        if (!tag.replayGainInfoIsEmpty()) {
                            ReplayGainInfo rg = tag.replayGainInfo();
                            appendMessage("ReplayGain info:",75,5,0);
                            if (!rg.albumGain.isEmpty()) appendMessage(QString("Album gain: %1").arg(rg.albumGain), 75, 10, 0);
                            if (!rg.albumPeak.isEmpty()) appendMessage(QString("Album peak: %1").arg(rg.albumPeak), 75, 10, 0);
                            if (!rg.albumMinMax.isEmpty()) appendMessage(QString("Album min/max: %1").arg(rg.albumMinMax), 75, 10, 0);
                            if (!rg.trackGain.isEmpty()) appendMessage(QString("Track gain: %1").arg(rg.trackGain), 75, 10, 0);
                            if (!rg.trackPeak.isEmpty()) appendMessage(QString("Track peak: %1").arg(rg.trackPeak), 75, 10, 0);
                            if (!rg.trackMinMax.isEmpty()) appendMessage(QString("Track min/max: %1").arg(rg.trackMinMax), 75, 10, 0);
                            if (!rg.loudness.isEmpty()) appendMessage(QString("Loudness: %1").arg(rg.loudness), 75, 10, 0);
                            if (!rg.undo.isEmpty()) appendMessage(QString("Undo: %1").arg(rg.undo), 75, 10, 0);
                        }


                        appendMessage("");
                    }
                    emit finished();
                }






                if (writeImageArg.isSet()) {
                    QString name = writeImageArg.getValue();
                    if (verbose) {
                        appendMessage("<font color=blue>Writing images:</font>");
                    }
                    bool ok=model.writeImage(name);
                    if (verbose && ok) {
                        showMessage(MT_SUCCESS, QString("Images were successfully written"));
                    }
                }

                if (scanRgArg.isSet() || writeRgArg.isSet()) {
                    QString typeS = scanRgArg.isSet() ? scanRgArg.getValue() : writeRgArg.getValue();
                    int type = RG_SCAN_AS_FILES;
                    if (typeS=="album") type=RG_SCAN_AS_ALBUM;
                    else if (typeS=="dir") type=RG_SCAN_AS_ALBUMS_BY_FOLDERS;
                    else if (typeS=="tag") type=RG_SCAN_AS_ALBUMS_BY_TAGS;
                    if (verbose) {
                        QString message="<font color=blue>Scanning ReplayGain info: ";
                        switch (type) {
                        case RG_SCAN_AS_FILES: message += "Scan per file track gain"; break;
                        case RG_SCAN_AS_ALBUM: message += "Scan as a single album"; break;
                        case RG_SCAN_AS_ALBUMS_BY_FOLDERS: message += "Scan as albums (by folder)"; break;
                        case RG_SCAN_AS_ALBUMS_BY_TAGS: message += "Scan as albums (by album tag)"; break;
                        }
                        appendMessage(message+"</font>");
                    }
                    if (!model.selectionSaved())
                        model.save(true);

                    ReplayGainer gainer(&model);
                    connect(&gainer,SIGNAL(message(int,QString)),this,SLOT(showMessage(int,QString)),Qt::QueuedConnection);
                    connect(&gainer,SIGNAL(textRead(QString)),SLOT(addText(QString)));
                    connect(&gainer,SIGNAL(tick()),SLOT(tick()));

                    gainer.setSkip(skipScannedArg.isSet());
                    gainer.start(type);
                    QList<ReplayGainInfo> newRg = gainer.getNewRgInfo();

                    //show rg
                    appendMessage("\n<font color=blue>ReplayGain scanning results</font>");
                    appendMessage("  Album gain  |  Track gain  |Album peak|Track peak| File name");
                    for (int i=0; i<model.selectionSize(); ++i) {
                        appendMessage(QString(" %1 | %2 | %3 | %4 | %5").arg(newRg.at(i).albumGain)
                                  .arg(newRg.at(i).trackGain)
                                  .arg(newRg.at(i).albumPeak)
                                  .arg(newRg.at(i).trackPeak)
                                  .arg(model.fileAtSelection(i).fileNameExt()));
                    }
                    appendMessage(" ");

                    if (writeRgArg.isSet() && !newRg.isEmpty()) {
                        //write rg
                        appendMessage("<font color=blue>Writing ReplayGain info...</font>");
                        for (int i=0; i<model.selectionSize(); ++i) {
                            Tag &tag = model.fileAtSelection(i);
                            tag.setReplayGainInfo(newRg.at(i));
                            tag.setChanged(true);
                        }
                        model.save();
                        appendMessage("<font color=green>Done.</font>");
                    }
                }

                if (removeRGArg.isSet()) {
                    if (verbose) {
                        appendMessage("<font color=blue>Removing ReplayGain info</font>");
                    }
                    if (!model.selectionSaved())
                        model.save(true);

                    ReplayGainer gainer(&model);
                    connect(&gainer,SIGNAL(message(QString)),this,SLOT(appendMessage(QString)),Qt::QueuedConnection);
                    gainer.setSkip(skipScannedArg.isSet());
                    gainer.start(RG_SCAN_REMOVE);
                }


                //removing tags
                if (removeAllArg.isSet()) {//removeAllTags, including standard tags, user tags, image, replaygain info
                    if (verbose) {
                        appendMessage("<font color=blue>Removing all tags:</font>");
                    }
                    model.removeAllTags();
                    if (verbose) {
                        showMessage(MT_SUCCESS, QString("Tags removing was successful"));
                    }
                }
                if (removeUserTagsArg.isSet()) {//remove all user tags
                    if (verbose) {
                        appendMessage("<font color=blue>Removing all user tags:</font>");
                    }
                    model.removeUserTags();
                    if (verbose) {
                        showMessage(MT_SUCCESS, QString("Tags removing was successful"));
                    }
                }
                if (removeArg.isSet()) {//remove standard tags by keys
                    bool success=true;
                    std::vector<QString> ids = removeArg.getValue();
                    for (uint i=0; i<ids.size(); ++i) {
                        QString key = ids[i];
                        if (key.isEmpty()) {
                            showMessage(MT_ERROR, QString("No key defined!"));
                            success=false;
                            continue;
                        }
                        int row = getRow(key);
                        if (verbose) {
                            appendMessage("<font color=blue>Removing tag with key \""+key+"\"</font>");
                        }
                        if (key>=0) model.setRow(row, "");
                        else model.setRow(key,"");
                    }
                    if (verbose && success) {
                        showMessage(MT_SUCCESS, QString("Tags removing was successful"));
                    }
                }
                if (removeImageArg.isSet()) {
                    if (verbose) {
                        appendMessage("<font color=blue>Removing image:</font>");
                    }
                    model.removeImage();
                    if (verbose) {
                        showMessage(MT_SUCCESS, QString("Image removing was successful"));
                    }
                }

                if (fillArg.isSet()) {
                    QString pattern = fillArg.getValue();
                    if (verbose) {
                        appendMessage("<font color=blue>Filling in tags from file name by \""+pattern+"\"</font>");
                    }
                    if (!pattern.isEmpty()) {
                        if (pattern.contains("%") && pattern.length()>=2) {
                            QStringList source=model.fileNames();
                            TagParser::updateSource(pattern,source);
                            QList<int> map;

                            QString placeholders = Placeholders::getWritablePlaceholders()
                                    + Placeholders::getVoidPlaceholder();
                            for (int i=0; i<model.selectionSize(); ++i) {
                                Tag &tag = model.fileAtSelection(i);
                                PairList parsed = TagParser::parse(source.at(i), pattern);
                                Q_FOREACH(const StringPair &pair, parsed) {
                                    int id = placeholders.indexOf(pair.first.at(1));
                                    if (id!=-1) {
                                        if (!map.contains(id)) map << id;
                                        tag.setTag(id, pair.second);
                                    }
                                }
                            }
                            if (verbose) {
                                showMessage(MT_SUCCESS, QString("Filling in tags was successful"));
                            }
                        }
                        else {
                            showMessage(MT_ERROR, QString("Fill pattern does not contain any placeholders"));
                        }
                    }
                    else {
                        showMessage(MT_ERROR, QString("Fill pattern is empty"));
                    }
                }

                //settings tags
                if (setArg.isSet()) {
                    std::vector<QString> ids = setArg.getValue();
                    bool success=true;
                    for (uint i=0; i<ids.size(); ++i) {
                        QString s = ids[i];
                        QString key = s.section(':',0,0);
                        QString value = s.section(':', 1);
                        if (key.isEmpty()) {
                            showMessage(MT_ERROR, QString("No key defined!"));
                            success=false;
                            continue;
                        }
                        int row = getRow(key);
                        if (verbose) {
                            appendMessage("<font color=blue>Changing tags with key \""+key+"\" to \""+value+"\":</font>");
                        }
                        if (key>=0) model.setRow(row, value);
                        else model.setRow(key,value);
                    }
                    if (verbose && success) {
                        showMessage(MT_SUCCESS, QString("Tags changing was successful"));
                    }
                }
                if (albumArg.isSet()) {
                    QString value = albumArg.getValue();
                    if (verbose) {
                        appendMessage("<font color=blue>Changing album to \""+value+"\":</font>");
                    }
                    model.setRow(ALBUM, value);
                    if (verbose) {
                        showMessage(MT_SUCCESS, QString("Album changing was successful"));
                    }
                }
                if (composerArg.isSet()) {
                    QString value = composerArg.getValue();
                    if (verbose) {
                        appendMessage("<font color=blue>Changing composer to \""+value+"\":</font>");
                    }
                    model.setRow(COMPOSER, value);
                    if (verbose) {
                        showMessage(MT_SUCCESS, QString("Composer changing was successful"));
                    }
                }
                if (titleArg.isSet()) {
                    QString value = titleArg.getValue();
                    if (verbose) {
                        appendMessage("<font color=blue>Changing title to \""+value+"\":</font>");
                    }
                    model.setRow(TITLE, value);
                    if (verbose) {
                        showMessage(MT_SUCCESS, QString("Title changing was successful"));
                    }
                }
                if (artistArg.isSet()) {
                    QString value = artistArg.getValue();
                    if (verbose) {
                        appendMessage("<font color=blue>Changing artist to \""+value+"\":</font>");
                    }
                    model.setRow(ARTIST, value);
                    if (verbose) {
                        showMessage(MT_SUCCESS, QString("Artist changing was successful"));
                    }
                }
                if (commentArg.isSet()) {
                    QString value = commentArg.getValue();
                    if (verbose) {
                        appendMessage("<font color=blue>Changing comment to \""+value+"\":</font>");
                    }
                    model.setRow(COMMENT, value);
                    if (verbose) {
                        showMessage(MT_SUCCESS, QString("Comment changing was successful"));
                    }
                    model.setRow(COMMENT, value);
                }
                if (genreArg.isSet()) {
                    QString value = genreArg.getValue();
                    if (verbose) {
                        appendMessage("<font color=blue>Changing genre to \""+value+"\":</font>");
                    }
                    model.setRow(GENRE, value);
                    if (verbose) {
                        showMessage(MT_SUCCESS, QString("Genre changing was successful"));
                    }
                }
                if (yearArg.isSet()) {
                    QString value = QString::number(yearArg.getValue());
                    if (verbose) {
                        appendMessage("<font color=blue>Changing year to \""+value+"\":</font>");
                    }
                    if (value=="0") value.clear();
                    model.setRow(YEAR, value);
                    if (verbose) {
                        showMessage(MT_SUCCESS, QString("Year changing was successful"));
                    }
                }
                if (trackNumberArg.isSet()) {
                    QString value = QString::number(trackNumberArg.getValue());
                    if (verbose) {
                        appendMessage("<font color=blue>Changing track number to \""+value+"\":</font>");
                    }
                    if (value=="0") value.clear();
                    model.setRow(TRACKNUMBER, value);
                    if (verbose) {
                        showMessage(MT_SUCCESS, QString("Track number changing was successful"));
                    }
                }
                if (totalTracksArg.isSet()) {
                    QString value = QString::number(totalTracksArg.getValue());
                    if (verbose) {
                        appendMessage("<font color=blue>Changing total tracks to \""+value+"\":</font>");
                    }
                    if (value=="0") value.clear();
                    model.setRow(TOTALTRACKS, value);
                    if (verbose) {
                        showMessage(MT_SUCCESS, QString("Total tracks changing was successful"));
                    }
                }
                if (setImageArg.isSet()) {
                    QString path = setImageArg.getValue();
                    if (verbose) {
                        appendMessage("<font color=blue>Inserting image from \""+path+"\":</font>");
                    }
                    if (path.isEmpty())
                        model.removeImage();
                    else if (QFile::exists(path)) {
                        QFile f(path);  f.open(QFile::ReadOnly);
                        QByteArray data = f.readAll();
                        QImage img;
                        if (img.loadFromData(data)) {
                            CoverImage cover(data,3,"image/png",f.fileName());
                            if (path.endsWith("jpg",Qt::CaseInsensitive) || path.endsWith("jpeg",Qt::CaseInsensitive))
                                cover.setMimetype("image/jpeg");
                            model.setImage(cover);
                            if (verbose) {
                                showMessage(MT_SUCCESS, QString("Image inserting was successful"));
                            }
                        }
                        else {
                            showMessage(MT_ERROR, "Cannot read image from \""+path+"\"");
                        }
                    }
                    else {
                        showMessage(MT_ERROR, "No file  \""+path+"\" exists");
                    }
                }

                if (setImageDescriptionArg.isSet()) {
                    QString description = setImageDescriptionArg.getValue();
                    if (verbose) {
                        appendMessage("<font color=blue>Setting image description to \""+description+"\":</font>");
                    }
                    model.setImageDescription(description);
                    if (verbose) {
                        showMessage(MT_SUCCESS, QString("Changing image description was successful"));
                    }
                }
                if (setImageTypeArg.isSet()) {
                    QString type = setImageTypeArg.getValue();
                    if (verbose) {
                        appendMessage("<font color=blue>Setting image type to \""+type+"\":</font>");
                    }
                    for (int i=0; i<21; ++i) {
                        if (type == imageTypes[i]) {
                            model.setImageType(i);
                            if (verbose) {
                                showMessage(MT_SUCCESS, QString("Changing image type was successful"));
                            }
                        }
                    }
                }
                model.save(true);

                if (renameArg.isSet()) {
                    QString pattern = renameArg.getValue();
                    if (verbose) {
                        appendMessage("<font color=blue>Renaming files with pattern \""+pattern+"\":</font>");
                    }
                    if (!pattern.isEmpty()) {
                        QStringList oldFileNames = model.fileNames();
                        FileNameRenderer filenameRenderer(&model);
                        filenameRenderer.setPattern(pattern);
                        QStringList newFileNames = filenameRenderer.getNewTags();
                        if (oldFileNames != newFileNames) {
                            model.rename(newFileNames);
                            if (verbose) {
                                showMessage(MT_SUCCESS, QString("New file names are:"));
                                Q_FOREACH(QString name, newFileNames) {
                                    appendMessage(name,75,5,0);
                                }
                            }
                        }
                        else
                            appendMessage("Nothing to rename, files names not changed with this pattern");
                    }
                    else {
                        showMessage(MT_ERROR, QString("Rename pattern is empty!"));
                    }
                }



                emit finished();
            }
            else {
                showMessage(MT_ERROR, QString("No files to process were specified!"));
                emit finished();
            }
        }
        else {
            new MainWindow();
        }
    }

    catch (TCLAP::ArgException &e)  {
        showMessage(MT_ERROR, e.error()+" for " + e.argId());
        emit finished();
    }
}

void ArgsParser::appendMessage(QString message, int maxWidth, int indentSpaces, int secondLineOffset)
{
    message.replace("<font color=red>","\033[31m");
    message.replace("<font color=green>","\033[32m");
    message.replace("<font color=blue>","\033[34m");
    message.replace("</font>","\033[0m");
    message.remove("<b>");
    message.remove("</b>");
    TCLAP::StdOutput::spacePrint(std::wcout, message, maxWidth, indentSpaces, secondLineOffset);
}

void ArgsParser::showMessage(int type, const QString &text)
{
    switch (type) {
    case MT_ERROR: appendMessage("<font color=red>Error: </font>"+text);
        break;
    case MT_INFORMATION: appendMessage(text);
        break;
    case MT_WARNING: appendMessage("<font color=red>Warning: </font>"+text);
        break;
    case MT_SUCCESS: appendMessage("<font color=green>OK</font>, "+text);
    }
}

void ArgsParser::addText(const QString &text)
{
    std::wcout << text.toStdWString() <<std::flush;
}

void ArgsParser::tick()
{
    //addText(".");
    std::wcout << "."<<std::flush;
}

void ArgsParser::printFiles(const QStringList &files)
{
    if (files.isEmpty()) return;
    showMessage(MT_SUCCESS, QString("%1 files were added:").arg(files.size()));
    Q_FOREACH(const QString &file, files) {
        appendMessage(file,75,5,5);
    }
    appendMessage("");
}
