/**
 * \file splitdialog.h
 * Dialog where you can split an image by cue file.
 *
 * \b Project: Qoobar
 * \author Alex Novichkov
 * \date 13 Sep 2010
 *
 * Copyright (C) 2009, 2010  Alex Novichkov
 *
 * This file is part of Qoobar.
 *
 * Qoobar is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Qoobar is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "splitdialog.h"
#include "application.h"
#ifdef HAVE_QT5
#include <QtWidgets>
#else
#include <QtGui>
#endif
#include "enums.h"

#include "cuesplitter.h"


SplitDialog::SplitDialog(QWidget *parent) :
    QDialog(parent)
{
    setWindowTitle(tr("Split an image file by cue"));

    warningLabel = new QLabel(this);
    warningLabel->setMaximumWidth(500);
    warningLabel->setOpenExternalLinks(true);

    cueFileEdit = new QLineEdit(this);
    cueFileEdit->setReadOnly(true);
    openCueButton = new QPushButton(tr("Open"),this);
    connect(openCueButton,SIGNAL(clicked()),SLOT(openCue()));

    folderEdit = new QLineEdit(this);
    folderEdit->setReadOnly(true);
    folderButton = new QPushButton(tr("Change"),this);
    connect(folderButton,SIGNAL(clicked()),SLOT(changeDir()));

    edit = new QPlainTextEdit(this);
    edit->setReadOnly(true);
    edit->setCenterOnScroll(true);
    edit->setMaximumBlockCount(1000);

    progress = new QProgressBar(this);
    progress->setMinimum(0);
   // progress->setEnabled(false);
    progress->setTextVisible(false);

    QDialogButtonBox *buttonBox = new QDialogButtonBox(Qt::Horizontal);
    startButton = buttonBox->addButton(tr("Start"), QDialogButtonBox::AcceptRole);
    QPushButton *cancelButton = buttonBox->addButton(QDialogButtonBox::Cancel);
    connect(startButton,SIGNAL(clicked()),SLOT(run()));
    connect(cancelButton,SIGNAL(clicked()),SLOT(cancel()));

    waitLabel = new QLabel(this);
    waitLabel->setMaximumWidth(500);
    waitLabel->hide();


    QGridLayout *grid = new QGridLayout;
    grid->addWidget(warningLabel,0,0,1,3);
    grid->addWidget(new QLabel(tr("Cue file:"),this),2,0);
    grid->addWidget(cueFileEdit,2,1);
    grid->addWidget(openCueButton,2,2);
    grid->addWidget(new QLabel(tr("Output folder:"),this),3,0);
    grid->addWidget(folderEdit,3,1);
    grid->addWidget(folderButton,3,2);
    grid->addWidget(edit,5,0,1,3);
    grid->addWidget(progress,6,0,1,3);
    grid->addWidget(waitLabel,7,0,1,3);
    grid->addWidget(buttonBox,9,0,1,3,Qt::AlignRight);
    grid->setRowMinimumHeight(1,10);
    grid->setRowMinimumHeight(4,10);
    grid->setRowMinimumHeight(8,10);

    setLayout(grid);
    warningLabel->setHidden(warningLabel->text().isEmpty());
    startButton->setEnabled(warningLabel->text().isEmpty());

    splitter = new CueSplitter(this);
    connect(splitter, SIGNAL(message(int,QString)),SLOT(onMessage(int,QString)),Qt::QueuedConnection);
    connect(splitter, SIGNAL(trackCountChanged(int)),progress,SLOT(setMaximum(int)),Qt::QueuedConnection);
    connect(splitter, SIGNAL(textReady(QString)),edit,SLOT(appendPlainText(QString)),Qt::QueuedConnection);
    connect(splitter, SIGNAL(next()), SLOT(onNextFile()),Qt::QueuedConnection);

    resize(600,400);
}

void SplitDialog::cancel()
{
    splitter->stop();
    reject();
}

void SplitDialog::changeDir()
{
    QString s=QFileDialog::getExistingDirectory(this,tr("Choose a folder to save files"),splitter->outputDir());
    if (!s.isEmpty()) {
        splitter->setOutputDir(s);
        folderEdit->setText(s);
    }
}

void SplitDialog::onNextFile()
{
    progress->setValue(progress->value()+1);
}

void SplitDialog::setCueFile(const QString &s)
{
    splitter->setCueFile(s);
    cueFileEdit->setText(s);
    folderEdit->setText(splitter->outputDir());
    waitLabel->hide();
}

void SplitDialog::openCue()
{
    QString s=QFileDialog::getOpenFileName(this, tr("Choose a cue file"),
                                           App->lastCueFile, QSL("*.cue"));
    if (!s.isEmpty()) setCueFile(s);
}

void SplitDialog::run()
{
    QApplication::setOverrideCursor(QCursor(Qt::WaitCursor));
    openCueButton->setEnabled(false);
    cueFileEdit->setEnabled(false);
    folderEdit->setEnabled(false);
    folderButton->setEnabled(false);
    startButton->setEnabled(false);

    splitter->split();
    splittedFiles = splitter->splittedFiles();

    QApplication::restoreOverrideCursor();
    progress->setValue(progress->maximum());

    if (!splittedFiles.isEmpty()) {
        accept();
    }
}


void SplitDialog::onMessage(int type, const QString &text)
{
    switch (type) {
    case MT_ERROR:
        warningLabel->show();
        warningLabel->setText(fontMetrics().elidedText(text,Qt::ElideRight,500));
        break;
    case MT_INFORMATION:
        waitLabel->show();
        waitLabel->setText(fontMetrics().elidedText(text,Qt::ElideRight,500));
        break;
    }
}
