/**
 * \file tablewidget.cpp
 * QTableWidget with additional events.
 *
 * \b Project: Qoobar
 * \author Alex Novichkov
 * \date 8 Dec 2009
 *
 * Copyright (C) 2009  Alex Novichkov
 *
 * This file is part of Qoobar.
 *
 * Qoobar is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Qoobar is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef HAVE_QT5
#include <QtWidgets>
#else
#include <QtGui>
#endif

#include "tablewidget.h"
#include "qoobarglobals.h"
#include "completerdelegate.h"
#include "application.h"
#include "displayedtagsdialog.h"



#include "enums.h"

TableWidget::TableWidget(QWidget* parent) : QTableWidget(TAGSCOUNT,1,parent),
    count(0)
{

    setContextMenuPolicy(Qt::ActionsContextMenu);
    setSelectionMode(QAbstractItemView::ExtendedSelection);
    setWordWrap(false);

#ifdef Q_OS_MAC
    setFrameStyle(QFrame::NoFrame | QFrame::Plain);
    //setAttribute(Qt::WA_MacShowFocusRect, false);
    setAutoFillBackground(true);
#endif

    for (int i=0; i<TAGSCOUNT; ++i) {
        setItem(i,0,new QTableWidgetItem(QString()));
        CompleterDelegate *d = new CompleterDelegate(i,false,this);
        setItemDelegateForRow(i,d);
    }
    viewport()->setAttribute(Qt::WA_Hover);

    horizontalHeader()->setStretchLastSection(true);
    horizontalHeader()->setVisible(false);
    setVerticalHeaderLabels(QVector<QString>(TAGSCOUNT,QString()).toList());
    if (!App->rows.isEmpty())
        verticalHeader()->restoreState(App->rows);
    verticalHeader()->SETSECTIONRESIZEMODE(QHeaderView::ResizeToContents);
    int height=verticalHeader()->sectionSize(0);
    verticalHeader()->SETSECTIONRESIZEMODE(QHeaderView::Fixed);
    verticalHeader()->setDefaultSectionSize(height);
    verticalHeader()->setContextMenuPolicy(Qt::ActionsContextMenu);
#ifdef HAVE_QT5
    verticalHeader()->setSectionsMovable(true);
#else
    verticalHeader()->setMovable(true);
#endif

    for (int i=0; i<TAGSCOUNT; ++i)
        verticalHeader()->setSectionHidden(i,App->tagStatus.at(i)==HIDDEN);

    QAction *a=new QAction(this);
    connect(a,SIGNAL(triggered()),this,SLOT(adjustDisplayedTags()));
    verticalHeader()->addAction(a);

    connect(verticalHeader(),SIGNAL(sectionMoved(int,int,int)),SLOT(updateTableHeaderGeometry()));
}

void TableWidget::retranslateUi()
{

    for (int i=0; i<TAGSCOUNT; ++i)
        verticalHeaderItem(i)->setText(tagByID(i));
    verticalHeader()->actions().at(0)->setText(tr("Adjust displayed tags..."));
}

void TableWidget::updateRow(int row, const QString &text)
{
    if (row<0 || row > rowCount()) return;
    item(row, 0)->setText(text);
    if (row<TAGSCOUNT)
        setRowHidden(row, App->tagStatus.at(row)==HIDDEN
                            || (App->tagStatus.at(row)==VISIBLEIFNOTEMPTY && text.isEmpty()));
}

void TableWidget::adjustDisplayedTags()
{
    DisplayedTagsDialog dialog(this);
    connect(&dialog,SIGNAL(tagStatusChanged(int,int)),SLOT(toggleTableRow(int,int)));
    dialog.exec();
}

void TableWidget::mouseDoubleClickEvent(QMouseEvent *event)
{

    if (!event || count==0) return;
    if (!itemAt(event->pos())) return;
    if (event->button()==Qt::LeftButton && count>1)
        emit cellClicked();
    else
        QTableWidget::mouseDoubleClickEvent(event);
}

void TableWidget::updateTableHeaderGeometry()
{

    App->rows=verticalHeader()->saveState();
}

void TableWidget::toggleTableRow(int row,int status)
{

    verticalHeader()->setSectionHidden(row,status==HIDDEN
                                       || (status==VISIBLEIFNOTEMPTY
                                           && item(row,0)->text().isEmpty()));
    App->tagStatus[row]=status;
    updateTableHeaderGeometry();
}

QString TableWidget::tag(int row) const
{

    return verticalHeaderItem(row)->text();
}

void TableWidget::addRow(const QString &header)
{

    int c=rowCount();
    blockSignals(true);
    setRowCount(c+1);
    setItem(c,0,new QTableWidgetItem(QString()));
    setVerticalHeaderItem(c,new QTableWidgetItem(header));
    blockSignals(false);
}

QList<int> TableWidget::selectedRows()
{

    QList<int> result;
    for (int i=0; i<rowCount(); ++i) {
        if (item(i,0)->isSelected()) result << i;
    }
    return result;
}
