#include "tagsreaderwriter.h"

#include "application.h"
#ifdef HAVE_QT5
#include <QtConcurrent/QtConcurrentMap>
#else
#include <QtConcurrentMap>
#endif
#include "stringroutines.h"

#include "taglib/mpeg/id3v2/frames/textidentificationframe.h"
#include "taglib/mpeg/id3v2/frames/unsynchronizedlyricsframe.h"
#include "taglib/mpeg/id3v2/frames/urllinkframe.h"
#include "taglib/mpeg/id3v2/frames/commentsframe.h"
#include "taglib/mpeg/id3v2/frames/attachedpictureframe.h"
#include "taglib/mpeg/id3v2/frames/popularimeterframe.h"
#include "taglib/mpc/mpcfile.h"
#include "taglib/ogg/flac/oggflacfile.h"
#include "taglib/trueaudio/trueaudiofile.h"
#include "taglib/ogg/speex/speexfile.h"
#include "taglib/wavpack/wavpackfile.h"
#include "taglib/mpeg/mpegfile.h"
#include "taglib/ogg/vorbis/vorbisfile.h"

#include "taglib/mpeg/id3v2/id3v2tag.h"
#include "taglib/ogg/xiphcomment.h"
#include "taglib/mpeg/id3v1/id3v1tag.h"
#include "taglib/ape/apetag.h"

#include "taglib/asf/asffile.h"
#include "taglib/riff/wav/wavfile.h"
#include "taglib/riff/aiff/aifffile.h"
#include "taglib/mp4/mp4file.h"
#include "taglib/mpeg/id3v2/frames/privateframe.h"
#include "taglib/mp4/mp4coverart.h"
#include "taglib/ogg/opus/opusfile.h"

#include "taglib/ape/apefile.h"
#include "taglib/ape/apeproperties.h"
#include "taglib/flac/flacfile.h"
#define APEFILE TagLib::APE::File

#define QS(s) QString::fromStdWString(s.toWString())
#define TS(s) TagLib::String(s.toStdWString())

const char id3v2tags[]=
"AENC.APIC.ASPI.COMM.COMR.ENCR.EQU2.ETCO.GEOB.GRID."
"LINK.MCDI.MLLT.OWNE.PRIV.PCNT.POPM.POSS.RBUF.RVA2."
"RVRB.SEEK.SIGN.SYLT.SYTC.TALB.TBPM.TCOM.TCON.TCOP."
"TDEN.TDLY.TDOR.TDRC.TDRL.TDTG.TENC.TEXT.TFLT.TIPL."
"TIT1.TIT2.TIT3.TKEY.TLAN.TLEN.TMCL.TMED.TMOO.TOAL."
"TOFN.TOLY.TOPE.TOWN.TPE1.TPE2.TPE3.TPE4.TPOS.TPRO."
"TPUB.TRCK.TRSN.TRSO.TSOA.TSOP.TSOT.TSRC.TSSE.TSST."
"TXXX.UFID.USER.USLT.WCOM.WCOP.WOAF.WOAR.WOAS.WORS."
"WPAY.WPUB.WXXX.EQUA.IPLS.RVAD.TDAT.TIME.TORY.TRDA."
"TSIZ.TYER.TCMP";

#include <QStringList>
#include <QTextCodec>
#include <QLibrary>
#include <QDir>

#include <qmath.h>

#include "enums.h"

TagsReaderWriter::TagsReaderWriter(): tag(0)
{
}

TagsReaderWriter::TagsReaderWriter(Tag *tag): tag(tag)
{
}

QString peak(int v)
{
    return QString::number((float)(qPow(10.f,(float)v / 5120.f)/32768.f),'f',6);
}
QString gain(int v)
{
    return QString("%2 dB").arg(v == 0 ? 0.f : 64.82f - v / 256.f,0,'f',2);
}

void TagsReaderWriter::readTags(int tagTypes)
{
    tag->clear();
    TagLib::File *f=readResolver(tagTypes);
    if (f) {
        TagLib::AudioProperties  *properties = f->audioProperties();
        if (properties) {
            tag->setLength(properties->length());
            tag->setBitrate(QString::number(properties->bitrate()));
            tag->setSampleRate(properties->sampleRate());
            tag->setChannels(properties->channels());

            TagLib::MPC::Properties *p=dynamic_cast<TagLib::MPC::Properties *>(properties);
            if (p) {
                ReplayGainInfo rg;
                if (p->trackPeak()!=0)
                    rg.trackPeak = peak(p->trackPeak());
                if (p->trackGain()!=0)
                    rg.trackGain = gain(p->trackGain());
                if (p->albumPeak()!=0)
                    rg.albumPeak = peak(p->albumPeak());
                if (p->albumGain()!=0)
                    rg.albumGain = gain(p->albumGain());
                if (!rg.isEmpty()) tag->setReplayGainInfo(rg);
            }
        }
        delete f;
    }

    static int intTags[] = {YEAR, TRACKNUMBER, TOTALTRACKS, RATING, DISCNUMBER, TOTALDISCS, TEMPO};
    for (int i=0; i<7; ++i) {
        if (tag->tag(intTags[i])=="0")
            tag->setTag1(intTags[i],QSL(""));
    }
}

TagLib::File *TagsReaderWriter::readResolver(int tagTypes)
{
    switch (tag->fileType()) {
    case Tag::MP3_FILE: {
        TagLib::MPEG::File *f=new TagLib::MPEG::File(FILE_NAME(tag->fullFileName()));
        if (!f) return 0;
        if (f->isValid()) {
            if (tagTypes<TAG_ALL) {//some of tags to read, not affected by app settings
                if (tagTypes & TAG_APE)   readAPE(f->APETag());
                if (tagTypes & TAG_ID3V2) readID3v2(f->ID3v2Tag());
                if (tagTypes & TAG_ID3V1) readID3v1(f->ID3v1Tag());
                return f;
            }
            if (App->mp3readape) readAPE(f->APETag());
            if (App->mp3readid3) readID3v2(f->ID3v2Tag());
            readID3v1(f->ID3v1Tag());
        }
        return f;
    }
    case Tag::TTA_FILE: {
        TagLib::TrueAudio::File *f=new TagLib::TrueAudio::File(FILE_NAME(tag->fullFileName()));
        if (!f) return 0;
        if (f->isValid()) {
            if (tagTypes & TAG_ID3V2) readID3v2(f->ID3v2Tag(false));
            if (tagTypes & TAG_ID3V1) readID3v1(f->ID3v1Tag());
        }
        return f;
    }
    case Tag::OGG_FILE: {
        TagLib::Ogg::Vorbis::File *f=new TagLib::Ogg::Vorbis::File(FILE_NAME(tag->fullFileName()));
        if (!f) return 0;
        if (f->isValid()) {
            if (tagTypes & TAG_VORBIS) readXiph(f->tag());
        }
        return f;
    }
    case Tag::FLAC_FILE: {
        TagLib::FLAC::File *f=new TagLib::FLAC::File(FILE_NAME(tag->fullFileName()));
        if (!f) return 0;
        if (f->isValid() && (tagTypes & TAG_VORBIS) ) {
            readXiph(f->xiphComment());
            readFlacPicture(f);
        }
        return f;
    }
    case Tag::OGA_FILE: {
        TagLib::Ogg::FLAC::File *f=new TagLib::Ogg::FLAC::File(FILE_NAME(tag->fullFileName()));
        if (f->isValid() && (tagTypes & TAG_VORBIS) ) {
            readXiph(f->tag());
            return f;
        }
        delete f;

        TagLib::Ogg::Vorbis::File *ff=new TagLib::Ogg::Vorbis::File(FILE_NAME(tag->fullFileName()));
        if (!ff) return 0;
        if (ff->isValid() && (tagTypes & TAG_VORBIS)) readXiph(ff->tag());
        return ff;
    }
    case Tag::SPX_FILE: {
        TagLib::Ogg::Speex::File *f=new TagLib::Ogg::Speex::File(FILE_NAME(tag->fullFileName()));
        if (!f) return 0;
        if (f->isValid() && (tagTypes & TAG_VORBIS)) readXiph(f->tag());
        return f;
    }
    case Tag::OPUS_FILE: {
        TagLib::Ogg::Opus::File *f=new TagLib::Ogg::Opus::File(FILE_NAME(tag->fullFileName()));
        if (!f) return 0;
        if (f->isValid() && (tagTypes & TAG_VORBIS)) readXiph(f->tag());
        return f;
    }
    case Tag::MPC_FILE: {
        TagLib::MPC::File *f=new TagLib::MPC::File(FILE_NAME(tag->fullFileName()));
        if (!f) return 0;
        if (f->isValid()) {
            if (tagTypes & TAG_APE) readAPE(f->APETag(false));
            if (tagTypes & TAG_ID3V1) readID3v1(f->ID3v1Tag());
        }
        return f;
    }
    case Tag::WV_FILE: {
        TagLib::WavPack::File *f=new TagLib::WavPack::File(FILE_NAME(tag->fullFileName()));
        if (!f) return 0;
        if (f->isValid()) {
            if (tagTypes & TAG_APE) readAPE(f->APETag(false));
            if (tagTypes & TAG_ID3V1) readID3v1(f->ID3v1Tag());
        }
        return f;
    }
    case Tag::APE_FILE: {
        APEFILE *f = new APEFILE(FILE_NAME(tag->fullFileName()));
        if (!f) return 0;
        if (f->isValid()) {
            if (tagTypes & TAG_APE) readAPE(f->APETag(false));
            if (tagTypes & TAG_ID3V1) readID3v1(f->ID3v1Tag());
        }
        return f;
    }

    case Tag::WAV_FILE: {
        TagLib::RIFF::WAV::File *f = new TagLib::RIFF::WAV::File(FILE_NAME(tag->fullFileName()));
        if (!f) return 0;
        if (f->isValid() && (tagTypes & TAG_ID3V2) )
            readID3v2(f->tag());
        return f;
    }
    case Tag::AIFF_FILE: {
        TagLib::RIFF::AIFF::File *f = new TagLib::RIFF::AIFF::File(FILE_NAME(tag->fullFileName()));
        if (!f) return 0;
        if (f->isValid() && (tagTypes & TAG_ID3V2) )
            readID3v2(f->tag());
        return f;
    }

    case Tag::WMA_FILE: {
        TagLib::ASF::File *f=new TagLib::ASF::File(FILE_NAME(tag->fullFileName()));
        if (!f) return 0;
        if (f->isValid() && (tagTypes & TAG_ASF) )
            readAsf(f->tag());
        return f;
    }

    case Tag::M4A_FILE: {
        TagLib::MP4::File *f=new TagLib::MP4::File(FILE_NAME(tag->fullFileName()));
        if (!f) return 0;
        if (f->isValid() && (tagTypes & TAG_MP4) )
            readMP4(f->tag());
        return f;
    }
    default: return 0;
    }
    return 0;
}

QStringList handleFrameList(const TagLib::ID3v2::FrameList &frames)
{
    QStringList result;
    TagLib::ID3v2::FrameList::ConstIterator it = frames.begin();
    for (; it!=frames.end(); ++it) {
        TagLib::String s=(*it)->toString();
        result.append(QS(s));
    }
    return result;
}

void TagsReaderWriter::readID3v2(TagLib::ID3v2::Tag *id3v2tag)
{
    if (!id3v2tag) return;

    const TagLib::ID3v2::FrameListMap &map = id3v2tag->frameListMap();

    TagLib::ID3v2::FrameListMap::ConstIterator it = map.begin();
    for (; it != map.end(); ++it)  {
        QString id = QS(TagLib::String((*it).first, TagLib::String::Latin1));

        if (id=="COMM") {
            for (unsigned int i=0; i<(*it).second.size(); ++i) {
                TagLib::ID3v2::CommentsFrame *frame=
                        dynamic_cast<TagLib::ID3v2::CommentsFrame *>((*it).second[i]);
                if (!frame) continue;
                parseTag(QString("COMM:%1").arg(QS(frame->description())), TaggingScheme::ID3, QS(frame->text()));
            }
        }
        else if (id=="PRIV") {
            for (unsigned int i=0; i<(*it).second.size(); ++i) {
                TagLib::ID3v2::PrivateFrame *frame=
                        dynamic_cast<TagLib::ID3v2::PrivateFrame *>((*it).second[i]);
                if (!frame) continue;

                QString owner = QS(frame->owner());
                //we skip all PRIV frames that we do not support
                //it is a bad decision, but I do not know how to
                //define whether PRIV frame contains text and not binary data
                if (owner == "WM/Mood") {
                    TagLib::String ts=TagLib::String(frame->data(),TagLib::String::UTF16LE);
                    parseTag(QString("PRIV:%1").arg(owner),TaggingScheme::ID3,QS(ts));
                }
            }
        }
        else if (id=="APIC") {
            if ((*it).second.size()>0) {
                TagLib::ID3v2::AttachedPictureFrame *frame=
                        dynamic_cast<TagLib::ID3v2::AttachedPictureFrame *>((*it).second[0]);
                if (!frame) continue;
                TagLib::ByteVector data=frame->picture();
                tag->setImageType(frame->type(), false);
                tag->setImageMimetype(QS(frame->mimeType()), false);
                tag->setImageDescription(QS(frame->description()), false);
                tag->setImagePixmap(QByteArray(data.data(),data.size()), false);
            }
        }
        else if (id=="POPM") {
            if ((*it).second.size()>0) {
                TagLib::ID3v2::PopularimeterFrame *frame=
                        dynamic_cast<TagLib::ID3v2::PopularimeterFrame *>((*it).second[0]);
                if (!frame) continue;
                int rating_=frame->rating();
                if (rating_>0) {
                    rating_ = (rating_-19)/59+1;
                    parseTag(QSL("POPM"),TaggingScheme::ID3,QString::number(rating_));
                }
            }
        }
        else if (id=="TCON")
            parseTag(id,TaggingScheme::ID3,QS(id3v2tag->genre()));
        else if (id=="TXXX") {
            QStringList values=handleFrameList((*it).second);
            Q_FOREACH (const QString &item, values) {
                QString idd = item.mid(1,item.indexOf(']')-1);
                parseTag(QString("%1:%2").arg(id).arg(idd),TaggingScheme::ID3,
                         item.right(item.length()-idd.length()*2-4));
            }
        }
        else if (id=="WXXX") {
            QStringList values=handleFrameList((*it).second);
            Q_FOREACH (const QString &item, values) {
                QString idd = item.mid(1,item.indexOf(']')-1);
                parseTag(QString("%1:%2").arg(id).arg(idd),TaggingScheme::ID3,
                         item.right(item.length()-idd.length()*2-3));
            }
        }
        else if (id.startsWith(QLS("T")) || id.startsWith(QLS("W"))) {
            QStringList values=handleFrameList((*it).second);
            parseTag(id,TaggingScheme::ID3,values.join(QSL(";")));
        }
        else if (id=="USLT") {
            QStringList values;
            for (unsigned int i=0; i<(*it).second.size(); ++i) {
                TagLib::ID3v2::UnsynchronizedLyricsFrame *frame=
                        dynamic_cast<TagLib::ID3v2::UnsynchronizedLyricsFrame *>((*it).second[i]);
                if (!frame) continue;
                values << QS(frame->text());
            }
            parseTag(id,TaggingScheme::ID3,values.join(QSL("\n")));
        }
    }
}

void parseRG(const QString &id, const QString &value, ReplayGainInfo &rg)
{
    if (id.contains(QSL("replaygain_album_gain"),Qt::CaseInsensitive))
        rg.albumGain = value;

    else if (id.contains(QSL("replaygain_album_peak"),Qt::CaseInsensitive))
        rg.albumPeak = value;

    else if (id.contains(QSL("replaygain_track_gain"),Qt::CaseInsensitive))
        rg.trackGain = value;

    else if (id.contains(QSL("replaygain_track_peak"),Qt::CaseInsensitive))
        rg.trackPeak = value;

    else if (id.contains(QSL("MP3GAIN_ALBUM_MINMAX"),Qt::CaseInsensitive))
        rg.albumMinMax = value;

    else if (id.contains(QSL("MP3GAIN_MINMAX"),Qt::CaseInsensitive))
        rg.trackMinMax = value;

    else if (id.contains(QSL("replaygain_album_minmax"),Qt::CaseInsensitive))
        rg.albumMinMax = value;

    else if (id.contains(QSL("replaygain_track_minmax"),Qt::CaseInsensitive))
        rg.trackMinMax = value;

    else if (id.contains(QSL("replaygain_reference_loudness"),Qt::CaseInsensitive))
        rg.loudness = value;

    else if (id.contains(QSL("MP3GAIN_UNDO"),Qt::CaseInsensitive))
        rg.undo = value;
}

//void TagsReaderWriter::readReplayGain(TagLib::ID3v2::Tag *id3v2tag)
//{
//    if (!id3v2tag) return;
//    ReplayGainInfo rg = tag->replayGainInfo();

//    const TagLib::ID3v2::FrameListMap &map = id3v2tag->frameListMap();
//    TagLib::ID3v2::FrameListMap::ConstIterator it = map.begin();
//    for (; it != map.end(); ++it) {
//        if ((*it).first=="TXXX") {
//            QStringList values=handleFrameList((*it).second);
//            Q_FOREACH (const QString &item, values) {
//                QString id = item.mid(1,item.indexOf(']')-1);
//                QString value=item.right(item.length()-id.length()*2-4);
//                if (!value.isEmpty()) parseRG(id, value, rg);
//            }
//        }
//    }
//    tag->setReplayGainInfo(rg, false);
//}

//void TagsReaderWriter::readReplayGain(TagLib::APE::Tag *apetag)
//{
//    if (!apetag) return;
//    ReplayGainInfo rg = tag->replayGainInfo();

//    TagLib::APE::ItemListMap map=apetag->itemListMap();
//    for (TagLib::APE::ItemListMap::ConstIterator it = map.begin(); it != map.end(); ++it)  {
//        if ((*it).second.type()!=0) continue;
//        QString id=QS((*it).first);
//        QString value=QS((*it).second.toString());
//        if (!value.isEmpty()) parseRG(id,value, rg);
//    }
//    tag->setReplayGainInfo(rg, false);
//}

void writeID3v2Frame(TagLib::ID3v2::Tag *id3v2tag,const QString &s,const QString &key)
{
    // f.e. id = "COMM" in "COMM:<description>"
    TagLib::ByteVector id = TagLib::ByteVector(key.toLatin1().data()).mid(0,4);
    QString description = key.mid(5);

    if (!QString(id3v2tags).contains(key.mid(0,4))) {
    //user text tags, write into TXXX frame
        id = "TXXX";
        description = key;
    }

    if (id=="COMM") { // comment
        QString comment1=s;
        comment1.replace(QLS("\n"),QLS("\r\n"));
        comment1.replace(QLS("\r\r"),QLS("\r"));

        if (description=="MusicMatch_Preference") {
            if (comment1=="1") comment1=QSL("Poor");
            else if (comment1=="2") comment1=QSL("Fair");
            else if (comment1=="3") comment1=QSL("Good");
            else if (comment1=="4") comment1=QSL("Very Good");
            else if (comment1=="5") comment1=QSL("Excellent");
        }
        //first we need to find comments
        TagLib::ID3v2::CommentsFrame *frame = TagLib::ID3v2::CommentsFrame::findByDescription(id3v2tag, TS(description));
        if (frame) {
            if (comment1.isEmpty()) id3v2tag->removeFrame(frame,true);
            else frame->setText(TS(comment1));
        }
        else if (!comment1.isEmpty()) {//create new
            frame = new TagLib::ID3v2::CommentsFrame(TagLib::String::UTF8);
            frame->setDescription(TS(description));
            //new frames get default language
            frame->setLanguage("eng");
            frame->setText(TS(comment1));
            id3v2tag->addFrame(frame);
        }
    }
    else if (id=="TXXX") {
        TagLib::ID3v2::UserTextIdentificationFrame *frame =
                TagLib::ID3v2::UserTextIdentificationFrame::find(id3v2tag, TS(description));
        if (frame)
            if (s.isEmpty()) id3v2tag->removeFrame(frame,true);
            else frame->setText(TS(s));
        else if (!s.isEmpty()) {
            frame = new TagLib::ID3v2::UserTextIdentificationFrame(TagLib::String::UTF8);
            frame->setDescription(TS(description));
            frame->setText(TS(s));
            id3v2tag->addFrame(frame);
        }
    }
    else if (id=="WXXX") {
        TagLib::ID3v2::UserUrlLinkFrame *frame =
                TagLib::ID3v2::UserUrlLinkFrame::find(id3v2tag, TS(description));
        if (frame)
            if (s.isEmpty()) id3v2tag->removeFrame(frame,true);
            else frame->setText(TS(s));
        else if (!s.isEmpty()) {
            frame = new TagLib::ID3v2::UserUrlLinkFrame(TagLib::String::Latin1);
            frame->setDescription(TS(description));
            frame->setUrl(TS(s));
            id3v2tag->addFrame(frame);
        }
    }
    else if (id=="PRIV") {
        TagLib::String data=TS(s);

        TagLib::ID3v2::PrivateFrame *frame = TagLib::ID3v2::PrivateFrame::find(id3v2tag,TS(description));
        if (frame)
            if (s.isEmpty()) id3v2tag->removeFrame(frame,true);
            else frame->setData(data.data(TagLib::String::UTF16LE));
        else if (!s.isEmpty()) {
            frame = new TagLib::ID3v2::PrivateFrame();
            frame->setOwner(TS(description));
            frame->setData(data.data(TagLib::String::UTF16LE));
            id3v2tag->addFrame(frame);
        }
    }
    else if (id=="USLT") {
        id3v2tag->removeFrames(id);
        if (!s.isEmpty()) {
            TagLib::ID3v2::UnsynchronizedLyricsFrame *frame =
                    new TagLib::ID3v2::UnsynchronizedLyricsFrame(TagLib::String::UTF8);
            frame->setText(TS(s));
            frame->setLanguage("eng");
            frame->setDescription("");
            id3v2tag->addFrame(frame);
        }
    }
    else if (id=="POPM") {
        id3v2tag->removeFrames(id);
        int rating_ = s.toInt();
        switch (rating_) {
        case 1: rating_ = 19; break;
        case 2: rating_ = 78; break;
        case 3: rating_ = 137; break;
        case 4: rating_ = 196; break;
        case 5: rating_ = 255; break;
        default: rating_ = 0; break;
        }
        if (rating_>0) {
            TagLib::ID3v2::PopularimeterFrame *popm=new TagLib::ID3v2::PopularimeterFrame();
            popm->setEmail(TagLib::String("Windows Media Player 9 Series",TagLib::String::Latin1));
            popm->setRating(rating_);
            popm->setCounter(0);
            id3v2tag->addFrame(popm);
        }
    }
    else if (id.startsWith("W")) {
        id3v2tag->removeFrames(id);
        if (!s.isEmpty()) {
            TagLib::ID3v2::UrlLinkFrame *frame = new TagLib::ID3v2::UrlLinkFrame(id);
            frame->setUrl(TS(s));
            id3v2tag->addFrame(frame);
        }
    }
    else if (id.startsWith("T")) {
        id3v2tag->removeFrames(id);
        if (!s.isEmpty()) {
            TagLib::ID3v2::TextIdentificationFrame *frame = new TagLib::ID3v2::TextIdentificationFrame(id,TagLib::String::UTF8);
            frame->setText(TS(s));
            id3v2tag->addFrame(frame);
        }
    }
    //remove empty frames but do not touch non-empty ones
    else if (s.isEmpty()) id3v2tag->removeFrames(id);
}

void TagsReaderWriter::writeID3v2(TagLib::ID3v2::Tag *id3v2tag)
{
    if (!id3v2tag) return;

    renderTags(TaggingScheme::ID3, id3v2tag);

    id3v2tag->setGenre(TS(tag->tag(GENRE)));

    id3v2tag->removeFrames("APIC");
    const CoverImage &image=tag->image();
    if (!image.pixmap().isEmpty()) {
        TagLib::ID3v2::AttachedPictureFrame *frame=
                new TagLib::ID3v2::AttachedPictureFrame();
        frame->setMimeType(TS(image.mimetype()));
        frame->setType((TagLib::ID3v2::AttachedPictureFrame::Type)image.type());
        frame->setDescription(TS(image.description()));
        frame->setPicture(TagLib::ByteVector(image.pixmap().data(),(unsigned int)image.pixmap().size()));
        id3v2tag->addFrame(frame);
    }

    //other tags
    const QMap<QString,QString> &otherTags = tag->userTags();
    QMap<QString,QString>::const_iterator i = otherTags.constBegin();
    for (; i != otherTags.constEnd(); ++i) {
        writeID3v2Frame(id3v2tag,i.value(),i.key());
    }

    //replaygain
    writeID3v2Frame(id3v2tag,tag->replayGainInfo().trackGain, QSL("TXXX:replaygain_track_gain"));
    writeID3v2Frame(id3v2tag,tag->replayGainInfo().trackPeak, QSL("TXXX:replaygain_track_peak"));
    writeID3v2Frame(id3v2tag,tag->replayGainInfo().albumGain, QSL("TXXX:replaygain_album_gain"));
    writeID3v2Frame(id3v2tag,tag->replayGainInfo().albumPeak, QSL("TXXX:replaygain_album_peak"));
    writeID3v2Frame(id3v2tag,tag->replayGainInfo().trackMinMax, QSL("TXXX:MP3GAIN_MINMAX"));
    writeID3v2Frame(id3v2tag,tag->replayGainInfo().albumMinMax, QSL("TXXX:MP3GAIN_ALBUM_MINMAX"));
    writeID3v2Frame(id3v2tag,tag->replayGainInfo().undo, QSL("TXXX:MP3GAIN_UNDO"));
    writeID3v2Frame(id3v2tag,tag->replayGainInfo().loudness, QSL("TXXX:replaygain_reference_loudness"));
}

void TagsReaderWriter::parseTag(const QString &id, const TaggingScheme::TagType tagType, const QString &v)
{
    if (v.isEmpty()) return;
    QString value=v;

    //hack for id3v2.4 COMM:MusicMatch_Preference
    if (id=="COMM:MusicMatch_Preference") {
        if (value=="Poor") value=QSL("1");
        else if (value=="Fair") value=QSL("2");
        else if (value=="Good") value=QSL("3");
        else if (value=="Very Good") value=QSL("4");
        else if (value=="Excellent") value=QSL("5");
    }

    if (id.contains(QSL("replaygain_album_gain"),Qt::CaseInsensitive)) {
        tag->setAlbumGain(value, false);
        return;
    }
    if (id.contains(QSL("replaygain_album_peak"),Qt::CaseInsensitive))  {
        tag->setAlbumPeak(value, false);
        return;
    }
    if (id.contains(QSL("replaygain_track_gain"),Qt::CaseInsensitive))  {
        tag->setTrackGain(value, false);
        return;
    }
    if (id.contains(QSL("replaygain_track_peak"),Qt::CaseInsensitive))  {
        tag->setTrackPeak(value, false);
        return;
    }
    if (id.contains(QSL("MP3GAIN_ALBUM_MINMAX"),Qt::CaseInsensitive))  {
        tag->setAlbumMinMax(value, false);
        return;
    }
    if (id.contains(QSL("MP3GAIN_MINMAX"),Qt::CaseInsensitive))  {
        tag->setTrackMinMax(value, false);
        return;
    }
    if (id.contains(QSL("replaygain_album_minmax"),Qt::CaseInsensitive)) {
        tag->setAlbumMinMax(value, false);
        return;
    }
    if (id.contains(QSL("replaygain_track_minmax"),Qt::CaseInsensitive)) {
        tag->setTrackMinMax(value, false);
        return;
    }
    if (id.contains(QSL("replaygain_minmax"),Qt::CaseInsensitive)) {
        tag->setTrackMinMax(value, false);
        return;
    }
    if (id.contains(QSL("replaygain_reference_loudness"),Qt::CaseInsensitive)) {
        tag->setLoudness(value, false);
        return;
    }
    if (id.contains(QSL("MP3GAIN_UNDO"),Qt::CaseInsensitive)) {
        tag->setReplayGainUndo(value, false);
        return;
    }

    if (id=="rtng") {//iTunes rating 1-100
        int r=value.toInt();
        value = QString::number(r / 20);
    }
    int idPosition = App->currentScheme->search(tagType,id);
    if (idPosition>=0) {
        if ((idPosition==TRACKNUMBER || idPosition==DISCNUMBER)
                && (App->currentScheme->equal(tagType,id,idPosition)) ) {
            QStringList l=value.split(QChar('/'));
            tag->appendTag(idPosition ,l.at(0));
            tag->appendTag(idPosition+1, l.value(1));
        }
        else if ((idPosition==TOTALTRACKS || idPosition==TOTALDISCS)
                 && (App->currentScheme->equal(tagType,id,idPosition-1))) {
            QStringList l=value.split('/');
            tag->appendTag(idPosition-1, l.at(0));
            tag->appendTag(idPosition, l.value(1));
        }
        else tag->appendTag(idPosition,value);
    }
    else tag->appendUserTag(id,value);
}

void TagsReaderWriter::readID3v1(TagLib::ID3v1::Tag *id3v1tag)
{
    if (!id3v1tag) return;

    if (tag->tag(ALBUM).isEmpty())
        tag->setTag1(ALBUM,QS(id3v1tag->album()));
    if (tag->tag(ARTIST).isEmpty())
        tag->setTag1(ARTIST,QS(id3v1tag->artist()));
    if (tag->tag(TITLE).isEmpty())
        tag->setTag1(TITLE,QS(id3v1tag->title()));
    if (tag->tag(GENRE).isEmpty())
        tag->setTag1(GENRE,QS(id3v1tag->genre()));
    if (tag->tag(COMMENT).isEmpty())
        tag->setTag1(COMMENT,QS(id3v1tag->comment()));
    if (tag->tag(YEAR).isEmpty() || tag->tag(YEAR)=="0")
            tag->setTag1(YEAR,QString::number(id3v1tag->year()));
    if (tag->tag(TRACKNUMBER).isEmpty() || tag->tag(TRACKNUMBER)=="0")
            tag->setTag1(TRACKNUMBER,QString::number(id3v1tag->track()));
}

void TagsReaderWriter::writeID3v1(TagLib::ID3v1::Tag *id3v1tag)
{
    if (!id3v1tag) return;

    if (App->id3v1Transliterate) {
        id3v1tag->setAlbum(TS(transliterate(tag->tag(ALBUM))));
        id3v1tag->setArtist(TS(transliterate(tag->tag(ARTIST))));
        id3v1tag->setComment(TS(transliterate(tag->tag(COMMENT))));
        id3v1tag->setGenre(TS(transliterate(tag->tag(GENRE))));
        id3v1tag->setTitle(TS(transliterate(tag->tag(TITLE))));
    }
    else {
        id3v1tag->setAlbum(TS(tag->tag(ALBUM)));
        id3v1tag->setArtist(TS(tag->tag(ARTIST)));
        id3v1tag->setComment(TS(tag->tag(COMMENT)));
        id3v1tag->setGenre(TS(tag->tag(GENRE)));
        id3v1tag->setTitle(TS(tag->tag(TITLE)));
    }
    id3v1tag->setTrack(tag->tag(TRACKNUMBER).toInt());
    id3v1tag->setYear(tag->tag(YEAR).toInt());
}

bool isFrontCover(TagLib::FLAC::Picture *p)
{
    return p->type()==TagLib::FLAC::Picture::FrontCover;
}

void TagsReaderWriter::readFlacPicture(TagLib::FLAC::File *f)
{
    TagLib::List<TagLib::FLAC::Picture *> pictures = f->pictureList();
    if (pictures.isEmpty()) return;

    TagLib::List<TagLib::FLAC::Picture *>::ConstIterator it = find_if(pictures.begin(),pictures.end(),isFrontCover);
    if (it==pictures.end())  it=pictures.begin();

    tag->setImageType((*it)->type(), false);
    tag->setImageMimetype(QS((*it)->mimeType()), false);
    tag->setImageDescription(QS((*it)->description()), false);
    tag->setImagePixmap(QByteArray((*it)->data().data(),(*it)->data().size()), false);
}

TagLib::FLAC::Picture *createFlacPicture(const CoverImage &image)
{
    TagLib::FLAC::Picture *picture = new TagLib::FLAC::Picture();
    QImage img;
    if (img.loadFromData(image.pixmap())) {
        picture->setData(TagLib::ByteVector(image.pixmap().data(),(unsigned int)image.pixmap().size()));
        picture->setDescription(TS(image.description()));
        picture->setMimeType(TS(image.mimetype()));
        picture->setType((TagLib::FLAC::Picture::Type)image.type());
        picture->setColorDepth(img.depth());
        picture->setHeight(img.height());
        picture->setWidth(img.width());
#if QT_VERSION >= 0x040600
        picture->setNumColors(img.colorCount());
#else
        picture->setNumColors(img.numColors());
#endif
    }
    return picture;
}

void TagsReaderWriter::writeFlacPicture(TagLib::FLAC::File *f)
{
    f->removePictures();
    if (!tag->image().pixmap().isEmpty()) {
        TagLib::FLAC::Picture *picture = createFlacPicture(tag->image());
        if (picture) f->addPicture(picture);
    }
}

void TagsReaderWriter::readXiph(TagLib::Ogg::XiphComment *xiph)
{
    if (!xiph) return;

    const TagLib::Ogg::FieldListMap &map=xiph->fieldListMap();

    for (TagLib::Ogg::FieldListMap::ConstIterator it = map.begin(); it != map.end(); ++it)  {
        QString id=QS((*it).first).toUpper();
        QString value=QS((*it).second.toString(";"));
        if (id=="METADATA_BLOCK_PICTURE") {
            QByteArray rawData = QByteArray::fromBase64(value.toLatin1());
            TagLib::ByteVector data=TagLib::ByteVector(rawData.data(),rawData.length());
            TagLib::FLAC::Picture pic(data);
            tag->setImageDescription(QS(pic.description()), false);
            tag->setImageType(pic.type(), false);
            tag->setImageMimetype(QS(pic.mimeType()), false);
            tag->setImagePixmap(QByteArray(pic.data().data(),pic.data().size()), false);
        }
        else if (id=="COVERARTMIME" && tag->fileType()!=Tag::FLAC_FILE)
            tag->setImageMimetype(value, false);
        else if (id=="COVERART" && tag->image().pixmap().isNull())
            tag->setImagePixmap(QByteArray::fromBase64(value.toLatin1()), false);
        else if (id=="COVERARTTYPE" && tag->fileType()!=Tag::FLAC_FILE)
            tag->setImageType(value.toInt(), false);
        else if (id=="COVERARTDESCRIPTION" && tag->fileType()!=Tag::FLAC_FILE)
            tag->setImageDescription(value, false);
        else parseTag(id,TaggingScheme::VORBIS,value);
    }
}

void TagsReaderWriter::writeXiph(TagLib::Ogg::XiphComment *xiph)
{
    if (!xiph) return;
    renderTags(TaggingScheme::VORBIS, xiph);

    xiph->removeField("COVERART");
    xiph->removeField("COVERARTMIME");
    xiph->removeField("COVERARTDESCRIPTION");
    xiph->removeField("COVERARTTYPE");
    xiph->removeField("METADATA_BLOCK_PICTURE");

    if (tag->fileType()!=Tag::FLAC_FILE && !tag->image().pixmap().isEmpty()) {
        if (App->oggPictureNew) {
            TagLib::FLAC::Picture *picture = createFlacPicture(tag->image());
            if (picture) {
                TagLib::ByteVector bv = picture->render();
                QByteArray result = QByteArray(bv.data(),bv.size());
                xiph->addField("METADATA_BLOCK_PICTURE",TS(QString(result.toBase64())),true);
            }
        }
        else {
            xiph->addField("COVERARTMIME",TS(tag->image().mimetype()),true);
            xiph->addField("COVERART",TS(QString(tag->image().pixmap().toBase64())),true);
            xiph->addField("COVERARTTYPE",TagLib::String::number(tag->image().type()),true);
            xiph->addField("COVERARTDESCRIPTION",TS(tag->image().description()),true);
        }
    }
    const QMap<QString, QString> &other = tag->userTags();
    QMap<QString, QString>::const_iterator it = other.constBegin();
    for (; it != other.constEnd(); ++it) {
        QStringList fields = App->writeFieldsSeparately ? it.value().split(QSL(";"),QString::SkipEmptyParts)
                                                        : QStringList(it.value()) ;

        xiph->removeField(TS(it.key().toUpper()));
        xiph->removeField(TS(it.key().toLower()));
        Q_FOREACH (const QString &field, fields)
            xiph->addField(TS(it.key()),TS(field),false);
    }

    xiph->addField("replaygain_track_gain",TS(tag->replayGainInfo().trackGain));
    xiph->addField("replaygain_track_peak",TS(tag->replayGainInfo().trackPeak));
    xiph->addField("replaygain_album_gain",TS(tag->replayGainInfo().albumGain));
    xiph->addField("replaygain_album_peak",TS(tag->replayGainInfo().albumPeak));
    xiph->addField("MP3GAIN_MINMAX",       TS(tag->replayGainInfo().trackMinMax));
    xiph->addField("MP3GAIN_ALBUM_MINMAX", TS(tag->replayGainInfo().albumMinMax));
    xiph->addField("MP3GAIN_UNDO",         TS(tag->replayGainInfo().undo));
    xiph->addField("replaygain_reference_loudness",TS(tag->replayGainInfo().loudness));
}

QStringList handleAttributeList(const TagLib::ASF::AttributeList &list)
{
    QStringList result;
    for (TagLib::ASF::AttributeList::ConstIterator it=list.begin(); it!=list.end(); ++it) {
        QString s=QS((*it).toString());
        if (!s.isEmpty()) result.append(s);
    }
    return result;
}

void TagsReaderWriter::readAsf(TagLib::ASF::Tag *asftag)
{
    if (!asftag) return;

    tag->setTag1(TITLE, QS(asftag->title()));
    tag->setTag1(ARTIST, QS(asftag->artist()));
    tag->setTag1(COPYRIGHT, QS(asftag->copyright()));
    int r = asftag->rating().toInt();
    if (r>0) tag->setTag1(RATING,QString::number(r));

    TagLib::ASF::AttributeListMap &map=asftag->attributeListMap();
    for (TagLib::ASF::AttributeListMap::ConstIterator it = map.begin(); it != map.end(); ++it)  {
        QString id=QS((*it).first);

        if (id=="WM/MediaPrimaryClassID" || id=="WMFSDKNeeded"
                || id=="WMFSDKVersion" || id=="DeviceConformanceTemplate")
            continue;
        if (id=="WM/Picture") {
            TagLib::ASF::Picture picture = (*it).second[0].toPicture();
            if (picture.isValid()) {
                tag->setImagePixmap(QByteArray(picture.picture().data(),picture.picture().size()), false);
                tag->setImageMimetype(QS(picture.mimeType()), false);
                tag->setImageDescription(QS(picture.description()), false);
                tag->setImageType((int)picture.type(), false);
            }
            continue;
        }
        QString value=handleAttributeList((*it).second).join(QSL(";"));
        parseTag(id,TaggingScheme::ASF,value);
    }
    // two more fields that are stored internally in wma, so read them if
    // specified fields are empty.
    if (tag->tag(COMMENT).isEmpty())
        tag->setTag1(COMMENT, QS(asftag->comment()));
    if (tag->tag(TRACKNUMBER).isEmpty())
        tag->setTag1(TRACKNUMBER, QString::number(asftag->track()));
}

void writeAsfItem(TagLib::ASF::Tag *tag,const QString &s,const TagLib::String &key)
{
    if (!s.isEmpty()) tag->addAttribute(key,TagLib::ASF::Attribute(TS(s)));
}

void TagsReaderWriter::writeAsf(TagLib::ASF::Tag *asftag)
{
    if (!asftag) return;

    renderTags(TaggingScheme::ASF, asftag);

    asftag->setArtist(TS(tag->tag(ARTIST)));
    asftag->setTitle(TS(tag->tag(TITLE)));
    asftag->setComment(TS(tag->tag(COMMENT)));
    asftag->setCopyright(TS(tag->tag(COPYRIGHT)));
    asftag->removeItem("WM/SharedUserRating");
    asftag->setRating(TagLib::String::number(qMin(tag->tag(RATING).toInt(),5)));
    //tag->removeItem("WM/BPM");

    if (!tag->image().pixmap().isEmpty()) {
        TagLib::ASF::Picture picture;
        picture.setDescription(TS(tag->image().description()));
        picture.setMimeType(TS(tag->image().mimetype()));
        picture.setPicture(TagLib::ByteVector(tag->image().pixmap().data(),tag->image().pixmap().size()));
        picture.setType((TagLib::ASF::Picture::Type)tag->image().type());
        asftag->setAttribute("WM/Picture",TagLib::ASF::Attribute(picture));
    } else asftag->removeItem("WM/Picture");

    const QMap<QString, QString> &other = tag->userTags();
    QMap<QString, QString>::const_iterator it = other.constBegin();
    for (; it != other.constEnd(); ++it) {
        QStringList fields = App->writeFieldsSeparately ? it.value().split(QSL(";"),QString::SkipEmptyParts)
                                                        : QStringList(it.value());
        asftag->removeItem(TS(it.key()));
        Q_FOREACH (const QString &field, fields)
            writeAsfItem(asftag,field,TS(it.key()));
    }
}

void TagsReaderWriter::readMP4(TagLib::MP4::Tag *mp4tag)
{
    if (!mp4tag)
        return;

    TagLib::MP4::ItemListMap map=mp4tag->itemListMap();

    for (TagLib::MP4::ItemListMap::ConstIterator it = map.begin(); it != map.end(); ++it)  {
        QString id=QS((*it).first);

        if (id=="trkn" || id=="disk") {
            int tn = (*it).second.toIntPair().first;
            int tt = (*it).second.toIntPair().second;
            parseTag(id,TaggingScheme::MP4,QString("%1/%2").arg(tn).arg(tt));
        }
        else if (id == "cpil" || id == "pgap" || id == "pcst" || id == "hdvd") {
            bool b=(*it).second.toBool();
            parseTag(id,TaggingScheme::MP4, b ? QSL("1") : QSL("0"));
        }
        else if (id=="tmpo") {
            parseTag(id, TaggingScheme::MP4, QString::number((*it).second.toInt()));
        }
        else if (id == "tvsn" || id == "tves" || id == "cnID" ||
                id == "sfID" || id == "atID" || id == "geID") {
            parseTag(id, TaggingScheme::MP4, QString::number((*it).second.toUInt()));
        }
        else if (id == "stik" || id == "rtng" || id == "akID") {
            parseTag(id, TaggingScheme::MP4, QString::number((*it).second.toByte()));
        }
        else if (id == "plID") {
            parseTag(id, TaggingScheme::MP4, QString::number((*it).second.toLongLong()));
        }
        else if (id == "rati") {
            parseTag(id, TaggingScheme::MP4, QString::number((*it).second.toUShort()));
        }
        else if (id=="covr") {
            TagLib::MP4::CoverArtList coverList=(*it).second.toCoverArtList();
            if (!coverList.isEmpty()) {
                TagLib::MP4::CoverArt cover=coverList[0];
                tag->setImageMimetype(cover.format()==TagLib::MP4::CoverArt::JPEG ? QSL("image/jpeg"):QSL("image/png"), false);
                tag->setImageDescription(QSL(""), false);
                TagLib::ByteVector data=cover.data();
                QByteArray bytes(data.data(),data.size());
                tag->setImagePixmap(bytes, false);
            }
        }
        else {
            QString value = QS((*it).second.toStringList().toString(";"));
            parseTag(id,TaggingScheme::MP4,value);
        }
    }
}

void writeMp4Item(TagLib::MP4::Tag *tag, const QString &id, const QString &val, bool other=false)
{
    if (!tag) return;
    TagLib::MP4::ItemListMap &map=tag->itemListMap();
    QString value = val;
    QString key=id;
    if (other)
        if (!key.startsWith(QLS("----:com.apple.iTunes:")))
            key.prepend(QLS("----:com.apple.iTunes:"));
    if (value.isEmpty()) {
        map.erase(TS(key));
        return;
    }
    if (key=="rtng") value = QString::number(value.toInt()*20);

    if (key=="trkn" || key=="disk") {
        QStringList list = value.split(QSL("/"));
        TagLib::MP4::Item pair(list.at(0).toInt(),list.value(1).toInt());
        map[TS(key)] = pair;
    }
    else if (key == "cpil" || key == "pgap" || key == "pcst" || key == "hdvd") {
        TagLib::MP4::Item item(value=="1" || value=="true");
        map[TS(key)] = item;
    }
    else if (key=="tmpo") {
        TagLib::MP4::Item item(value.toInt());
        map[TS(key)] = item;
    }
    else if (key == "tvsn" || key == "tves" || key == "cnID" ||
            key == "sfID" || key == "atID" || key == "geID") {
        TagLib::MP4::Item item(value.toUInt());
        map[TS(key)] = item;
    }
    else if (key == "stik" || key == "rtng" || key == "akID") {
        TagLib::MP4::Item item(value.toUShort());
        map[TS(key)] = item;
    }
    else if (key == "plID") {
        TagLib::MP4::Item item(value.toLongLong());
        map[TS(key)] = item;
    }
    else if (key == "rati") {
        TagLib::MP4::Item item(value.toUShort()); item.setAtomDataType(TagLib::MP4::TypeImplicit);
        map[TS(key)] = item;
    }
    else {
         map[TS(key)] = TagLib::StringList(TS(value));
    }
}

void TagsReaderWriter::writeMP4(TagLib::MP4::Tag *mp4tag)
{
    if (!mp4tag) return;

    TagLib::MP4::ItemListMap &map=mp4tag->itemListMap();
    renderTags(TaggingScheme::MP4,mp4tag);

    map.erase("covr");
    if (!tag->image().pixmap().isEmpty()) {
        TagLib::MP4::CoverArt::Format format=TagLib::MP4::CoverArt::JPEG;
        if (tag->image().mimetype()=="image/png") format=TagLib::MP4::CoverArt::PNG;
        TagLib::MP4::CoverArt cover(format,TagLib::ByteVector(tag->image().pixmap().data(),(unsigned int)tag->image().pixmap().size()));

        TagLib::MP4::CoverArtList l;
        l.append(cover);
        map["covr"]=TagLib::MP4::Item(l);
    }

    const QMap<QString, QString> &other = tag->userTags();
    QMap<QString, QString>::const_iterator it = other.constBegin();
    for (; it != other.constEnd(); ++it) {
        writeMp4Item(mp4tag,it.key(),it.value(),true);
    }

    const ReplayGainInfo &rg = tag->replayGainInfo();
    writeMp4Item(mp4tag,QSL("replaygain_track_gain"),rg.trackGain,true);
    writeMp4Item(mp4tag,QSL("replaygain_track_peak"),rg.trackPeak,true);
    writeMp4Item(mp4tag,QSL("replaygain_album_gain"),rg.albumGain,true);
    writeMp4Item(mp4tag,QSL("replaygain_album_peak"),rg.albumPeak,true);
    writeMp4Item(mp4tag,QSL("replaygain_track_minmax"),rg.trackMinMax,true);
    writeMp4Item(mp4tag,QSL("replaygain_album_minmax"),rg.albumMinMax,true);
    writeMp4Item(mp4tag,QSL("replaygain_mp3gain_undo"),rg.undo,true);
    writeMp4Item(mp4tag,QSL("replaygain_reference_loudness"),rg.loudness,true);

    mp4tag->save();
}

void TagsReaderWriter::readAPE(TagLib::APE::Tag *apetag)
{
    if (!apetag) return;

    TagLib::APE::ItemListMap map=apetag->itemListMap();

    for (TagLib::APE::ItemListMap::ConstIterator it = map.begin(); it != map.end(); ++it)  {
        QString id=QS((*it).first).toUpper();

        if (id=="COVER ART (FRONT)") {
            TagLib::ByteVector v=(*it).second.value();
            QByteArray data=QByteArray(v.data(),v.size());
            int pos=data.indexOf('\0',0);
            int pos1=data.lastIndexOf('.',pos);
            QByteArray descr=data.mid(0,pos);
            tag->setImageDescription(QString::fromUtf8(descr.constData()), false);
            QByteArray mime=data.mid(pos1+1,pos-pos1-1);
            tag->setImageMimetype(QSL("image/png"), false);
            if (mime.toUpper()=="JPG" || mime.toUpper()=="JPEG")
                tag->setImageMimetype(QSL("image/jpeg"), false);
            tag->setImagePixmap(data.mid(pos+1), false);
            tag->setImageType(3, false);
        }
        else if ((*it).second.type()==TagLib::APE::Item::Text) {
            QString value=QS((*it).second.toString());
            parseTag(id,TaggingScheme::APE,value);
        }
    }
}

void TagsReaderWriter::writeAPE(TagLib::APE::Tag *apetag)
{
    if (!apetag) return;

    renderTags(TaggingScheme::APE,apetag);
    apetag->removeItem("Cover Art (front)");
    QByteArray bytes=tag->image().pixmap();
    if (!bytes.isEmpty()) {
        bytes.prepend((char)0x0);
        if (tag->image().description().isEmpty()) {
            if (tag->image().mimetype()=="image/jpeg")
                bytes.prepend("front.jpg");
            else
                bytes.prepend("front.png");
        }
        else
            bytes.prepend(tag->image().description().toUtf8());

        TagLib::ByteVector data;
        data.append(TagLib::ByteVector::fromUInt(bytes.size(), false));
        data.append(TagLib::ByteVector::fromUInt(2, false));
        data.append("Cover Art (front)");
        data.append(TagLib::ByteVector('\0'));
        data.append(TagLib::ByteVector(bytes.constData(),bytes.size()));

        TagLib::APE::Item item;
        item.parse(data);
        apetag->setItem("Cover Art (front)",item);
    }

    const QMap<QString, QString> &other = tag->userTags();
    QMap<QString, QString>::const_iterator it = other.constBegin();
    for (; it != other.constEnd(); ++it)
        apetag->setItem(TS(it.key()),TagLib::APE::Item(TS(it.key()),TS(it.value())));

    //replaygain
    if (tag->fileType()==Tag::MPC_FILE && App->mpcWriteRg)
        tag->removeReplayGainInfo(false);
    else {
        const ReplayGainInfo &rg = tag->replayGainInfo();
        apetag->addValue("replaygain_track_gain",TS(rg.trackGain));
        apetag->addValue("replaygain_track_peak",TS(rg.trackPeak));
        apetag->addValue("replaygain_album_gain",TS(rg.albumGain));
        apetag->addValue("replaygain_album_peak",TS(rg.albumPeak));
        apetag->addValue("MP3GAIN_MINMAX",       TS(rg.trackMinMax));
        apetag->addValue("MP3GAIN_ALBUM_MINMAX", TS(rg.albumMinMax));
        apetag->addValue("MP3GAIN_UNDO",         TS(rg.undo));
        apetag->addValue("replaygain_reference_loudness",TS(rg.loudness));
    }
}


void TagsReaderWriter::renderTags(const TaggingScheme::TagType type, TagLib::Tag *taglibtag)
{
    for (int i=0; i<TAGSCOUNT; ++i) {
        QStringList fields = App->currentScheme->fields(i,type);

        Q_FOREACH (const QString &field, fields) {
            QString value = tag->tag(i);
            if (i==RATING) {
                int r=qMin(value.toInt(),5);
                value=r>0?QString::number(r) : QSL("");
            }
            if ((i==TRACKNUMBER || i==DISCNUMBER)
                    && App->currentScheme->equal(type,field,i)
                    && !tag->tag(i+1).isEmpty()) {
                value = QString("%1/%2").arg(value).arg(tag->tag(i+1));
            }
            else if ((i==TOTALTRACKS || i==TOTALDISCS)
                     && App->currentScheme->equal(type,field,i-1)) {
                continue;
            }
            switch (type) {
            case TaggingScheme::VORBIS: {
                TagLib::Ogg::XiphComment *t =  static_cast<TagLib::Ogg::XiphComment *>(taglibtag);
                QStringList tags = App->writeFieldsSeparately ? value.split(QSL(";"),QString::SkipEmptyParts) : QStringList(value);
                t->removeField(TS(field));

                Q_FOREACH (const QString &v, tags) {
                    t->addField(TS(field),TS(v),false);
                }
                break;
            }
            case TaggingScheme::APE: {
                TagLib::APE::Tag *t = static_cast<TagLib::APE::Tag *>(taglibtag);
                if (t) {
                    t->setItem(TS(field),TagLib::APE::Item(TS(field),TS(value)));
                }
                break;
            }
            case TaggingScheme::ID3:
                writeID3v2Frame(static_cast<TagLib::ID3v2::Tag *>(taglibtag) ,value, field);
                break;
            case TaggingScheme::ASF: {
                TagLib::ASF::Tag *t = static_cast<TagLib::ASF::Tag *>(taglibtag);
                QStringList tags = App->writeFieldsSeparately ?  value.split(QSL(";"),QString::SkipEmptyParts) : QStringList(value);
                t->removeItem(TS(field));
                Q_FOREACH (const QString &v, tags) {
                     writeAsfItem(t,v,TS(field));
                }
                break;
            }
            case TaggingScheme::MP4: {
                TagLib::MP4::Tag *mp4= static_cast<TagLib::MP4::Tag *>(taglibtag);
                writeMp4Item(mp4,field,value);
                break;
            }
            default: break;
            }
        }
    }
}

bool TagsReaderWriter::writeTags()
{
    if (!tag->wasChanged())
        return true;
    bool b=true;

    if (App->trim) {
        for (int i=0; i<TAGSCOUNT; ++i) {
            if (i!=COMMENT && i!=LYRICS) tag->setTag1(i, tag->tag(i).simplified());
        }
        QMap<QString, QString> other = tag->userTags();
        QMutableMapIterator<QString, QString> i(other);
        while (i.hasNext()) {
            i.next();
            i.setValue(i.value().simplified());
        }
        tag->setUserTags1(other);
    }

    QString lyrics = tag->tag(LYRICS);
    lyrics.replace(QLS("\n"),QLS("\r\n"));
    lyrics.replace(QLS("\r\r"),QLS("\r"));
    tag->setTag1(LYRICS,lyrics);

    if (tag->image().mimetype().isEmpty())
        tag->setImageMimetype(QSL("image/png"), false);
    if (tag->image().type() < 0 || tag->image().type() > 0x14)
        tag->setImageType(3, false);



    switch (tag->fileType()) {
    case Tag::MP3_FILE: {
        TagLib::MPEG::File *f=new TagLib::MPEG::File(FILE_NAME(tag->fullFileName()));
        if (f && f->isValid()) {
            int savingTypes=0;
            if (App->mp3writeape) {
                TagLib::APE::Tag *tag=f->APETag(true);
                if (tag) {
                    writeAPE(tag);
                    savingTypes |=TagLib::MPEG::File::APE;
                }
            }
            if (App->mp3writeid3) {
                TagLib::ID3v2::Tag *tag=f->ID3v2Tag(true);
                if (tag) {
                    writeID3v2(tag);
                    savingTypes |= TagLib::MPEG::File::ID3v2;
                }
            }
            if (App->id3v1Synchro<2) {
                TagLib::ID3v1::Tag *tag=f->ID3v1Tag(App->id3v1Synchro==0);
                if (tag) {
                    writeID3v1(tag);
                    savingTypes |= TagLib::MPEG::File::ID3v1;
                }
            }
            b=f->save(savingTypes,true,App->id3v2version);
        }
        delete f;
        break;
    }
    case Tag::TTA_FILE: {
        TagLib::TrueAudio::File *f=new TagLib::TrueAudio::File(FILE_NAME(tag->fullFileName()));
        if (f && f->isValid()) {
            TagLib::ID3v2::Tag *tag=f->ID3v2Tag(true);
            writeID3v2(tag);

            if (App->id3v1Synchro<2) {//do write id3v1
                TagLib::ID3v1::Tag *tag=f->ID3v1Tag(App->id3v1Synchro==0);
                writeID3v1(tag);
            }
            else f->strip(TagLib::TrueAudio::File::ID3v1);
            b=f->save();
        }
        delete f;
        break;
    }
    case Tag::OGG_FILE: {
        TagLib::Ogg::Vorbis::File *f=new TagLib::Ogg::Vorbis::File(FILE_NAME(tag->fullFileName()));
        if (f && f->isValid()) {
            TagLib::Ogg::XiphComment *tag=f->tag();
            writeXiph(tag);
            b=f->save();
        }
        delete f;
        break;
    }
    case Tag::OPUS_FILE: {
        TagLib::Ogg::Opus::File *f=new TagLib::Ogg::Opus::File(FILE_NAME(tag->fullFileName()));
        if (f && f->isValid()) {
            TagLib::Ogg::XiphComment *tag=f->tag();
            writeXiph(tag);
            b=f->save();
        }
        delete f;
        break;
    }
    case Tag::FLAC_FILE: {
        TagLib::FLAC::File *f = new TagLib::FLAC::File(FILE_NAME(tag->fullFileName()));
        if (f && f->isValid()) {
            TagLib::Ogg::XiphComment *tag=f->xiphComment(true);
            writeXiph(tag);
            writeFlacPicture(f);
            b=f->save();
        }
        delete f;
        break;
    }
    case Tag::OGA_FILE: {
        TagLib::Ogg::FLAC::File *f = new TagLib::Ogg::FLAC::File(FILE_NAME(tag->fullFileName()));
        if (f && f->isValid()) {
            TagLib::Ogg::XiphComment *tag=f->tag();
            writeXiph(tag);
            b=f->save();
        }
        delete f;
        TagLib::Ogg::Vorbis::File *ff=new TagLib::Ogg::Vorbis::File(FILE_NAME(tag->fullFileName()));
        if (ff && ff->isValid()) {
            TagLib::Ogg::XiphComment *tag=ff->tag();
            writeXiph(tag);
            b=ff->save();
        }
        delete ff;
        break;
    }
    case Tag::SPX_FILE: {
        TagLib::Ogg::Speex::File *f=new TagLib::Ogg::Speex::File(FILE_NAME(tag->fullFileName()));
        if (f && f->isValid()) {
            TagLib::Ogg::XiphComment *tag=f->tag();
            writeXiph(tag);
            b=f->save();
        }
        delete f;
        break;
    }
    case Tag::MPC_FILE: {
        TagLib::MPC::File *f=new TagLib::MPC::File(FILE_NAME(tag->fullFileName()));
        if (f && f->isValid()) {
            TagLib::APE::Tag *tag=f->APETag(true);
            writeAPE(tag);
            if (App->id3v1Synchro<2) {//do write id3v1
                TagLib::ID3v1::Tag *tag=f->ID3v1Tag(App->id3v1Synchro==0);
                writeID3v1(tag);
            }
            else f->strip(TagLib::MPC::File::ID3v1);
            b=f->save();
        }
        delete f;
        break;
    }
    case Tag::WV_FILE: {
        TagLib::WavPack::File *f=new TagLib::WavPack::File(FILE_NAME(tag->fullFileName()));
        if (f && f->isValid()) {
            TagLib::APE::Tag *tag=f->APETag(true);
            writeAPE(tag);

            if (App->id3v1Synchro<2) {//do write id3v1
                TagLib::ID3v1::Tag *tag=f->ID3v1Tag(App->id3v1Synchro==0);
                writeID3v1(tag);
            }
            else f->strip(TagLib::WavPack::File::ID3v1);
            b=f->save();
        }
        delete f;
        break;
    }
    case Tag::WAV_FILE: {
        TagLib::RIFF::WAV::File *f = new TagLib::RIFF::WAV::File(FILE_NAME(tag->fullFileName()));
        if (f && f->isValid()) {
            TagLib::ID3v2::Tag *tag=f->tag();
            writeID3v2(tag);
            b=f->save();
        }
        delete f;
        break;
    }

    case Tag::WMA_FILE: {
        TagLib::ASF::File *f=new TagLib::ASF::File(FILE_NAME(tag->fullFileName()));
        if (f && f->isValid()) {
            TagLib::ASF::Tag *tag=f->tag();
            writeAsf(tag);
            b=f->save();
        }
        delete f;
        break;
    }

    case Tag::AIFF_FILE: {
        TagLib::RIFF::AIFF::File *f = new TagLib::RIFF::AIFF::File(FILE_NAME(tag->fullFileName()));
        if (f && f->isValid()) {
            TagLib::ID3v2::Tag *tag=f->tag();
            writeID3v2(tag);
            b=f->save();
        }
        delete f;
        break;
    }

    case Tag::M4A_FILE: {
        TagLib::MP4::File *f=new TagLib::MP4::File(FILE_NAME(tag->fullFileName()));
        if (f && f->isValid()) {
            TagLib::MP4::Tag *tag=f->tag();
            writeMP4(tag);
            b=f->save();
        }
        delete f;
        break;
    }
    case Tag::APE_FILE: {
        APEFILE *f=new APEFILE(FILE_NAME(tag->fullFileName()));
        if (f && f->isValid()) {
            TagLib::APE::Tag *tag=f->APETag(true);
            writeAPE(tag);
            if (App->id3v1Synchro<2) {//do write id3v1
                TagLib::ID3v1::Tag *tag=f->ID3v1Tag(App->id3v1Synchro==0);
                writeID3v1(tag);
            }
            else f->strip(1);
            b=f->save();
        }
        delete f;
        break;
    }
    }
    tag->setChanged(!b);
    return b;
}
